'''
This module provides options(), which returns a argparse.Namespace object
containing all configurations required by sepp. config.options() reads and
saves the configurations the first time it is called. Subsequent calls simply
return the already-saved configurations. A Typical usage is:
some_config_attribute = sepp.config.options().some_config_attribute

All command-line options are found directly inside the return value of
options() as attributes. Commandline options can be specified either directly
using the command-line, or under [commandline] section in a config file,
passed to SEPP using -c commandline option. commandline values overwrite
config file values.

Information stored in other sections of the config file are available as
nested arparse.Namespaes attributes inside the results of options(),
with the config file header used as the attribute name. For example,
imagine the config file has:
[pplacer]
path = /some/path
In this case, config.options().pplacer will be a arparse.Namespace and
config.options().pplacer.path will be "/some/path".

A "main" configuration file under {home}/.sepp/main.config is used
to store some basic configurations such as the location of extra programs, etc.
This main config file is read first, so that user provided config file can
overwrite its values.

In addition, each client of this module (e.g. a new algorithm)) can add new
commandline options by getting the parser object using get_parser() and then
adding extra options. This has to happen *before* the first call to options()
module. For an example see exhaustive_upp.
'''

from argparse import ArgumentParser, Namespace
from sepp.filemgr import get_default_temp_dir, check_or_make_dir_path
import sys
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
from sepp import version, get_logger
import argparse
import os
import os.path
from multiprocessing import cpu_count
from sepp import scheduler
import random

_LOG = get_logger(__name__)

root_p = open(os.path.join(os.path.split(
    os.path.split(__file__)[0])[0], "home.path")).readlines()[0].strip()
main_config_path = os.path.join(root_p, "main.config")


def set_main_config_path(filename):
    global main_config_path
    main_config_path = filename


def _read_config_file(filename, opts, expand=None):
    config_defaults = []
    cparser = configparser.ConfigParser()
    cparser.optionxform = str
    cparser.readfp(filename)

    if cparser.has_section('commandline'):
        for (k, v) in cparser.items('commandline'):
            config_defaults.append("--%s" % k)
            config_defaults.append(v)

    for section in cparser.sections():
        if section == "commandline":
            continue
        section_name_space = Namespace()
        for (k, v) in cparser.items(section):
            if expand and k == "path":
                v = os.path.join(expand, v)
            section_name_space.__setattr__(k, v)
        opts.__setattr__(section, section_name_space)

    return config_defaults


def valid_dir_path(path):
    ret = check_or_make_dir_path(path)
    if ret is None:
        raise argparse.ArgumentTypeError(
            "%s is not a valid directory path." % path)
    return ret


def valid_molecule(molecule):
    ret = molecule in ['dna', 'rna', 'amino']
    if ret is False:
        raise argparse.ArgumentTypeError(
            ("%s is not a valid molecule type.  Must be 'dna', 'rna', or "
             "'amino'.") % molecule)
    return molecule


def valid_decomp_strategy(strategy):
    ret = strategy in ['hierarchical', 'normal', 'midpoint', 'centroid']
    if ret is False:
        raise argparse.ArgumentTypeError(
            ("%s is not a valid strategy.  Must be 'normal', 'hierarchical',"
             " 'centroid', or 'midpoint'.") % strategy)
    return strategy


def valid_file_prefix(prefix):
    if os.path.dirname(prefix) != "":
        raise argparse.ArgumentTypeError(
            "%s is not a valid output prefix (includes a directory)." % prefix)
    return prefix


def set_cpu(cpus):
    c = int(cpus)
    scheduler.default_cpus = c
    return c


def set_checkpoint(checkpoint):
    import sepp.checkpointing
    return sepp.checkpointing.CheckPointManager(checkpoint)


_parser = None


def _init_parser():
    global _parser
    _parser = ArgumentParser(
        description=(
            "This script runs the SEPP algorithm on an input "
            "tree, alignment, fragment file, and RAxML info file."),
        conflict_handler='resolve')

    _parser.add_argument("-v", "--version", action='version',
                         version="%(prog)s " + version)

    decompGroup = _parser.add_argument_group(
        "Decomposition Options".upper(),
        ' '.join([
            "These options determine the alignment decomposition size and",
            "taxon insertion size.  If None is given, then the default",
            "is to align/place at 10% of total taxa.  "
            "The alignment decomosition size must be",
            "less than the taxon insertion size."]))
    _parser.groups = dict()
    _parser.groups['decompGroup'] = decompGroup

    decompGroup.add_argument(
        "-A", "--alignmentSize", type=int,
        dest="alignment_size", metavar="N", default=None,
        help=("max alignment subset size of N "
              "[default: 10%% of the total number of taxa or the placement"
              " subset size if given]"))
    decompGroup.add_argument(
        "-P", "--placementSize", type=int,
        dest="placement_size", metavar="N", default=None,
        help=("max placement subset size of N "
              "[default: 10%% of the total number of taxa or the alignment "
              "length (whichever bigger)]"))
    decompGroup.add_argument(
        "-F", "--fragmentChunkSize", type=int,
        dest="max_chunk_size", metavar="N", default=20000,
        help=("maximum fragment chunk size of N. Helps controlling memory.  "
              "[default: 20000]"))
    decompGroup.add_argument(
        "-D", "--distance", type=float,
        dest="distance", metavar="DISTANCE",
        default=1,
        help=("minimum p-distance before stopping the decomposition"
              "[default: 1]"))
# uym2 added #
    decompGroup.add_argument(
        "-M", "--diameter", type=float,
        dest="maxDiam", metavar="DIAMETER",
        default=None,
        help=("maximum tree diameter before stopping the decomposition"
              "[default: None]"))

    decompGroup.add_argument(
        "-S", "--decomp_strategy", type=valid_decomp_strategy,
        dest="decomp_strategy", metavar="DECOMP",
        default="normal",
        # default = "midpoint",
        help="decomposition strategy "
        "[default: using tree branch length]")
    # "[default: only include smallest subsets]")

    outputGroup = _parser.add_argument_group(
        "Output Options".upper(), "These options control output.")
    _parser.groups['outputGroup'] = outputGroup

    outputGroup.add_argument(
        "-p", "--tempdir",
        dest="tempdir", metavar="DIR",
        type=valid_dir_path,
        default=get_default_temp_dir(),
        help=("Tempfile files will be written to DIR. Full-path required. "
              "[default: %(default)s]"))
    outputGroup.add_argument(
        "-o", "--output",
        dest="output", metavar="OUTPUT",
        default="output",
        type=valid_file_prefix,
        help="output files with prefix OUTPUT. [default: %(default)s]")
    outputGroup.add_argument(
        "-d", "--outdir",
        dest="outdir", metavar="OUTPUT_DIR",
        default=os.path.curdir,
        type=valid_dir_path,
        help=("output to OUTPUT_DIR directory. full-path required. "
              "[default: %(default)s]"))

    inputGroup = _parser.add_argument_group(
        "Input Options".upper(),
        ' '.join([
            "These options control input. To run SEPP the following is "
            "required. A backbone tree (in newick format), a RAxML_info file "
            "(this is the file generated by RAxML during estimation of the "
            "backbone tree. Pplacer uses this info file to set model "
            "parameters), a backbone alignment file (in fasta format), and a"
            " fasta file including fragments.  The input sequences are assumed"
            " to be DNA unless specified otherwise."]))
    _parser.groups['inputGroup'] = inputGroup

    inputGroup.add_argument(
        "-c", "--config",
        dest="config_file", metavar="CONFIG",
        type=argparse.FileType('r'),
        help=("A config file, including options used to run SEPP. Options"
              " provided as command line arguments overwrite config file "
              "values for those options. "
              "[default: %(default)s]"))
    inputGroup.add_argument(
        "-t", "--tree",
        dest="tree_file", metavar="TREE",
        type=argparse.FileType('r'),
        help="Input tree file (newick format) [default: %(default)s]")
    inputGroup.add_argument(
        "-r", "--raxml",
        dest="info_file", metavar="RAXML",
        type=argparse.FileType('r'),
        help=("RAxML_info file including model parameters, generated by RAxML."
              "[default: %(default)s]"))
    inputGroup.add_argument(
        "-a", "--alignment",
        dest="alignment_file", metavar="ALIGN",
        type=argparse.FileType('r'),
        help="Aligned fasta file [default: %(default)s]")
    inputGroup.add_argument(
        "-f", "--fragment",
        dest="fragment_file", metavar="FRAG",
        type=argparse.FileType('r'),
        help="fragment file [default: %(default)s]")
    inputGroup.add_argument(
        "-m", "--molecule",
        dest="molecule", metavar="MOLECULE",
        type=valid_molecule,
        default="dna",
        help=("Molecule type of sequences. Can be amino, dna, or rna "
              "[default: %(default)s]"))

    otherGroup = _parser.add_argument_group(
        "Other options".upper(), "These options control how SEPP is run")
    _parser.groups['otherGroup'] = otherGroup
    otherGroup.add_argument(
        "-x", "--cpu", type=set_cpu,
        dest="cpu", metavar="N",
        default=set_cpu(cpu_count()),
        help=("Use N cpus "
              "[default: number of cpus available on the machine]"))
    otherGroup.add_argument(
        "-cp", "--checkpoint", type=set_checkpoint,
        dest="checkpoint", metavar="CHCK_FILE",
        default=set_checkpoint(None),
        help="checkpoint file [default: no checkpointing]")
    otherGroup.add_argument(
        "-cpi", "--interval", type=int,
        dest="checkpoint_interval", metavar="N",
        default=3600,
        help=("Interval (in seconds) between checkpoint writes. Has effect "
              "only with -cp provided. [default: 3600]"))
    otherGroup.add_argument(
        "-seed", "--randomseed", type=int,
        dest="seed", metavar="N",
        default=297834,
        help="random seed number. [default: 297834]")
    # inputGroup.add_argument("-p", "--package",
    #                  dest="package", metavar="PKG",
    #                  help="package directory"
    #                         "[default: %(default)s]")
    #

    return _parser


def get_parser():
    global _parser
    if _parser is None:
        _parser = _init_parser()
    return _parser


def _parse_options():
    parser = get_parser()

    opts = Namespace()

    ''' First read the main configuration file '''
    _LOG.debug("Main configuration file at %s" % main_config_path)
    if not os.path.exists(main_config_path):
        _LOG.warn(
            "Main configuration file was not found at: %s\n" %
            main_config_path + "Proceeding without the main configuration...")
        main_cmd_defaults = []
    else:
        main_cmd_defaults = _read_config_file(
            open(main_config_path, 'r'), opts)

    input_args = main_cmd_defaults + (sys.argv[1:])

    ''' Then read the commandline options '''
    opts = parser.parse_args(input_args, namespace=opts)

    ''' If there is a user-specified config file, read that '''
    if opts.config_file is not None:

        config_cmd_defaults = _read_config_file(opts.config_file, opts)

        input_args = main_cmd_defaults + config_cmd_defaults + (sys.argv[1:])

        def error_callback(message):
            newmessage = message.replace(
                "arguments:",
                "arguments (potentially from the config file):").replace(
                    "--", "")
            ArgumentParser.error(parser, newmessage)

        parser.error = error_callback

        ''' Read commandline options again to overwrite config file values'''
        opts = parser.parse_args(input_args, namespace=opts)
    random.seed(opts.seed)
    _LOG.info("Seed number: %d" % opts.seed)
    return opts


_options_singelton = None


def options():
    '''
    Returns the configurations read from main configuration file,
    commandline and the user input configuration file.
    '''
    global _options_singelton
    if _options_singelton is None:
        _options_singelton = _parse_options()
    return _options_singelton



'''
Created on Oct 2, 2012

@author: smirarab
'''
from sepp.config import options
from abc import abstractmethod
from sepp.scheduler import JobPool
from sepp.filemgr import directory_has_files_with_prefix, get_temp_file
from sepp.alignment import MutableAlignment
from sepp.tree import PhylogeneticTree
import dendropy
from sepp import get_logger
import sys
import os
from sepp.problem import SeppProblem
import time
from sepp.checkpointing import CheckPointManager

_LOG = get_logger(__name__)


class AbstractAlgorithm(object):
    '''
    This class provides the interface for an abstract algorithm. All different
    ways of implementing SEPP are considered to be different 'algorithms'.

    New algorithms should be implemented by extending AbastractAlgorithm and
    implementing its abstract methods.
    '''

    def __init__(self):
        '''
        Constructor
        '''
        self.root_problem = None
        self.results = None
        self.options = options()
        self.outchecked = False  # for ease of access

    def check_options(self, supply=[]):
        '''
        This method should check the input values stored in config.option to
        make sure every necessary argument is provided, and that the provided
        values are all fine.

        In the event of recognizing invalid or missing input, an Exception
        (maybe an ArgumentError) should be raised.

        By default expects tree_file, raxml_file, and fragment_file. Overwrite
        if required.
        '''
        if (options().tree_file is None):
            supply = supply + ["tree file"]
        if (options().alignment_file is None):
            supply = supply + ["alignment file"]
        if (options().fragment_file is None):
            supply = supply + ["fragment file"]
        if (len(supply) != 0):
            raise ValueError(
                ("Failed to supply: %s\nRun with -h option to see a list of"
                 " options.")
                % " , ".join(supply))
        self.check_outputprefix()

    @abstractmethod
    def build_subproblems(self):
        '''
        This method should read the config.options() and build a hierarchy of
        subproblems (see sepp.problem). This hierarchy will be used by
        build_job_dag to create a DAG of jobs (see sepp.jobs and
        sepp.scheduler)
        '''
        raise NotImplementedError()

    @abstractmethod
    def connect_jobs(self):
        '''
        Join Jobs are joined together to form a DAG using Joins (see
        sepp.scheduler)
        and call_back functions (see sepp.scheduler.Job).

        Once the first level of jobs (those with no dependency) are enqueued
        (using enqueue_firstlevel_job) everything else should be automatically
        enqueued when all their dependencies are satisfied.
        For example, if the conceptual DAG is:
        A----> E---|--> F
                   |
        B--|-> D---|
        C--|
        then enqueue_firstlevel_job should enqueue A B and C. When A finishes,
        it should enqueue E (either using a callback or a Join with only one
        dependency). B and C should be joined together using a Join, and that
        Join needs to enqueue D. Also E and D need to be joined, and their join
        needs to enqueue F.
        build_jobs is called before this, and all jobs are saved as part of the
        subproblem hierarchy. This function only connects those jobs.
        This is called after recovery from a checkpoint, because joins and
        callbacks are not checkpointed.
        '''
        raise NotImplementedError()

    @abstractmethod
    def build_jobs(self):
        '''
        Builds separate jobs for different tasks that need to be done.
        This just creates jobs, without connecting them. connect_jobs is used
        to create jobs.
        build_job is not called after checkpoints are recovered, because the
        jobs are checkpointed (their connections are not).
        '''
        raise NotImplementedError()

    @abstractmethod
    def enqueue_firstlevel_job(self):
        '''
        This is called after the DAG is created (see buld_job_dag) to enqueue
        the first level of jobs (those with no dependency). These jobs should
        automatically enqueue the rest of the DAG upon completion.
        '''
        raise NotImplementedError()

    @abstractmethod
    def merge_results(self):
        '''
        This method should read results and prepare all the final results
        for output. Resutls should be saved in the problem hierarchy, and
        by setting ?? attribute
        '''
        raise NotImplementedError()

    @abstractmethod
    def output_results(self):
        '''
        This method should output the final results. Final results can be found
        in the problem hierarchy
        '''
        raise NotImplementedError()

    def run(self):
        checkpoint_manager = options().checkpoint
        assert isinstance(checkpoint_manager, CheckPointManager)

        t = time.time()

        if checkpoint_manager.is_recovering:
            checkpoint_manager.restore_checkpoint()
            self.root_problem = \
                checkpoint_manager.checkpoint_state.root_problem
            self.check_outputprefix()
        else:
            '''check input arguments'''
            self.check_options()

            '''build the problem structure'''
            self.root_problem = self.build_subproblems()

            '''build jobs'''
            self.build_jobs()

        '''connect jobs into a DAG'''
        self.connect_jobs()

        '''Queue up first level jobs (i.e. those with no dependency).
        Once these run, they should automatically enqueue the rest of the
        DAG through joins and callbacks '''
        self.enqueue_firstlevel_job()

        '''start the checkpointing (has any effects only in
           checkpointing mode)'''
        checkpoint_manager.start_checkpointing(self.root_problem)

        '''Wait for all jobs to finish'''
        if (not JobPool().wait_for_all_jobs()):
            _LOG.exception(
                "There have been errors in executed jobs. Terminating.")
            sys.exit(1)

        ''' terminate The job pool and release memory'''
        JobPool().terminate()

        ''' Pause Checkpointing'''
        checkpoint_manager.pause_checkpointing()
        # checkpoint_manager.force_checkpoint()

        '''Merge results into final outputs'''
        self.merge_results()

        '''Output final results'''
        self.output_results()

        ''' Pause Checkpointing'''
        checkpoint_manager.stop_checkpointing()

        _LOG.info("Current execution Finished in %d seconds"
                  % (time.time() - t))
        _LOG.info(
            "All checkpointed executions Finished in %d cumulative time" %
            (checkpoint_manager.get_total_time()))

    ''' The following are a bunch of helper methods that will be needed in
    most implementations of sepp'''

    def check_and_set_sizes(self, total):
        # If sizes are not set, then use 10% rule
        options = self.options
        if (options.maxDiam is not None) and \
           (options.decomp_strategy not in ["midpoint", "centroid"]):
            raise Exception(("The max diameter option can be used only with "
                             "the midpoint or the centroid decomposition "
                             "specified using -S"))
        if (options.alignment_size is None):
            if options.placement_size is None:
                options.alignment_size = int(total*.10)
            else:
                options.alignment_size = options.placement_size
        if (options.placement_size is None):
            options.placement_size = max(int(total * .10),
                                         options.alignment_size)
        if options.placement_size is not None and \
           options.placement_size < options.alignment_size:
            _LOG.warning(
                ("alignment_size (%d) cannot be larger than placement_size "
                 "(%d).   Setting alignment_size to be placement_size") %
                (options.alignment_size, options.placement_size))
            options.alignment_size = options.placement_size
        if options.placement_size > total:
            options.placement_size = total
            _LOG.warning(
                ("Input placement size larger than total number of sequences!"
                 "  Setting placement size to %d") % total)
        if options.alignment_size > total:
            options.alignment_size = total
            _LOG.warning(
                ("Input alignment size larger than total number of sequences!"
                 "  Setting alignment size to %d") % total)

        _LOG.info(
            "Decomposition Sizes are set to alignment: %d placement: %d" %
            (options.alignment_size, options.placement_size))

    def check_outputprefix(self):
        if self.outchecked:
            return
        self.outchecked = True
        if directory_has_files_with_prefix(
           self.options.outdir, self.options.output):
            raise ValueError(
                ("Output directory [%s] already contains files with prefix "
                 "[%s]...\nTerminating to avoid loss of existing files.") % (
                    self.options.outdir, self.options.output))

    def get_output_filename(self, name):
        return os.path.join(
            self.options.outdir, "%s_%s" % (self.options.output, name))

    def read_alignment_and_tree(self):
        _LOG.info("Reading input alignment: %s" % (
            self.options.alignment_file))
        alignment = MutableAlignment()
        alignment.read_file_object(self.options.alignment_file)

        # fragments = MutableAlignment()
        # fragments.read_file_object(self.options.fragment_file);
        _LOG.info("Reading input tree: %s" % self.options.tree_file)
        tree = PhylogeneticTree(
            dendropy.Tree.get_from_stream(self.options.tree_file,
                                          schema="newick",
                                          preserve_underscores=True))

        return (alignment, tree)

    def read_and_divide_fragments(self, chunks, extra_frags={}):
        max_chunk_size = self.options.max_chunk_size
        _LOG.debug(
            ("start reading fragment files and breaking to at least %s chunks"
             " but at most %s sequences ") % (
                str(chunks), str(max_chunk_size)))
        self.root_problem.fragments = MutableAlignment()
        self.root_problem.fragments.read_file_object(
            self.options.fragment_file)

        # test if input fragment names might collide with reference names.
        # code contribution by Stefan Janssen (June 13th, 2018)
        ids_reference = set(self.root_problem.subalignment.keys())
        ids_inputfragments = set(self.root_problem.fragments.keys())
        ids_overlap = ids_reference & ids_inputfragments
        if len(ids_overlap) > 0:
            raise ValueError((
                "Your input fragment file contains %i sequences, whose names "
                "overlap with names in your reference. Please rename your inp"
                "ut fragments and re-start. Duplicate names are:\n  '%s'") %
                (len(ids_overlap), "'\n  '".join(ids_overlap)))

        # test if input fragment names contain whitespaces / tabs which would
        # cause hmmsearch to fail.
        # code contribution by Stefan Janssen (June 22nd, 2018)
        ids_inputfragments_spaces = [
            id_
            for id_ in ids_inputfragments
            if (' ' in id_) or ('\t' in id_)]
        if len(ids_inputfragments_spaces) > 0:
            raise ValueError((
                "Your input fragment file contains %i sequences, whose names "
                "contain either whitespaces: ' ' or tabulator '\\t' symbols. "
                "Please rename your input fragments and re-start. Affected "
                "names are:\n  '%s'") %
                (len(ids_inputfragments_spaces),
                 "'\n  '".join(ids_inputfragments_spaces)))

        for (k, v) in extra_frags.items():
            self.root_problem.fragments[k] = v.replace("-", "")
        alg_chunks = self.root_problem.fragments.divide_to_equal_chunks(
            chunks, max_chunk_size)
        ret = []
        for i in range(0, len(alg_chunks)):
            temp_file = None
            if alg_chunks[i]:
                temp_file = get_temp_file(
                    "fragment_chunk_%d" % i, "fragment_chunks", ".fasta")
                alg_chunks[i].write_to_path(temp_file)
            ret.append(temp_file)
        _LOG.debug("fragment files read and divided.")
        return ret

    def _create_root_problem(self, tree, alignment):
        ''' Create the root problem'''
        self.root_problem = SeppProblem(tree.leaf_node_names())
        self.root_problem.label = "root"
        self.root_problem.subalignment = alignment
        self.root_problem.subtree = tree



###########################################################################
#    Copyright 2012 Siavash Mirarab, Nam Nguyen, and Tandy Warnow.
#    This file is part of SEPP.
#
#    SEPP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SEPP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SEPP.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################

from operator import itemgetter
import logging
import os


__all__ = ["alignment", "shortreadalignment", "taxonneighbourfinder",
           "tools", "problem"]
version = "4.3.10"
_DEBUG = ("SEPP_DEBUG" in os.environ) and \
    (os.environ["SEPP_DEBUG"].lower() == "true")

_INSTALL_PATH = __path__[0]


def is_temp_kept():
    return _DEBUG


def get_setup_path():
    return _INSTALL_PATH


def get_logging_level():
    return logging.DEBUG if _DEBUG else logging.INFO


__set_loggers = set()


def get_logger(name="sepp"):
    logger = logging.getLogger(name)
    if name not in __set_loggers:
        level = get_logging_level()
        logging_formatter = logging.Formatter(
            ("[%(asctime)s] %(filename)s (line %(lineno)d):"
             " %(levelname) 8s: %(message)s"))
        logging_formatter.datefmt = '%H:%M:%S'
        logger.setLevel(level)
        ch = logging.StreamHandler()
        ch.setLevel(level)
        ch.setFormatter(logging_formatter)
        logger.addHandler(ch)
        __set_loggers.add(name)
    return logger


def reset_loggers():
    global __set_loggers
    __set_loggers = set()
    import pkgutil
    import sepp
    for l, name, _ in pkgutil.iter_modules(['sepp']):
        logger = (getattr(getattr(sepp, name, None), "_LOG", None))
        if logger:
            setattr(getattr(sepp, name, None), "_LOG", get_logger(
                "sepp.%s" % name))


def log_exception(logger):
    '''Logs the exception trace to the logObj as an error'''
    import traceback
    import io
    s = io.StringIO()
    traceback.print_exc(None, s)
    logger.debug(s.getvalue())


os.sys.setrecursionlimit(1000000)


def sortByValue(d, reverse=False):
    return sorted(iter(d.items()), key=itemgetter(1), reverse=reverse)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from sepp.exhaustive import ExhaustiveAlgorithm

# #########################################################################, a#
#    Copyright 2012 Siavash Mirarab, Nam Nguyen, and Tandy Warnow.
#    This file is part of SEPP.
#
#    SEPP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SEPP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SEPP.  If not, see <http://www.gnu.org/licenses/>.
# ##########################################################################


# MAIN
if __name__ == '__main__':
    ExhaustiveAlgorithm().run()



"""SATe - Phylogenetic Tree Container, effectively a wrapper of
   dendropy.Tree"""

# This file is part of SATe

# SATe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Jiaye Yu and Mark Holder, University of Kansas

# This file is copied to SEPP and is used as an external library

from dendropy import Tree, Taxon, treecalc
from dendropy import DataSet as Dataset
from dendropy.datamodel.treemodel import _convert_node_to_root_polytomy as \
    convert_node_to_root_polytomy
from sepp import get_logger, sortByValue
from sepp.alignment import get_pdistance
from sepp.decompose_tree import decompose_by_diameter


try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import copy
import string
import random
import re

_LOG = get_logger(__name__)


class PhylogeneticTree(object):
    """Data structure to store phylogenetic tree, wrapping dendropy.Tree."""
    def __init__(self, dendropy_tree, map_internal_node_names=True):
        self._tree = dendropy_tree
        assert isinstance(self._tree, Tree)
        self.n_leaves = self.count_leaves()
        self._tree.seed_node.edge.tail_node = None
        self._tree.seed_node.edge.length = None
        self._namemap = None
        self._revscript = None

        if map_internal_node_names:
            self.map_seq_names()

    def map_seq_names(self):
        i = 1
        tag = ''.join(random.choice(string.ascii_letters)
                      for letter in range(8))
        self._namemap = dict()
        revnamemap = dict()
        for n in self._tree.internal_nodes():
            if n.label:
                l1 = '%sN%06d%s' % (tag, i, re.sub(
                    r"[%s]" % string.punctuation.replace("_", " "),
                    repl="_", string=n.label))
                self._namemap[n] = l1
                revnamemap[l1] = n.label
                i = i + 1
        self._revscript = '''import ast, re, string, sys
revnamemap = ast.literal_eval("%s")
def relabel_newick(newick_string):
    pattern = re.compile("(%sN[^(,:)<>]+)")
    invalidChars = set(string.punctuation).union(set(string.whitespace))
    def replace_func(m):
        repl = m.group(1)
        if m.group(1) in revnamemap:
            repl = revnamemap[m.group(1)]
            if any(char in invalidChars for char in repl):
                repl = "'%%s'" %%repl
        else:
            repl = m.group(1)

        return repl
    t = pattern.sub(replace_func,newick_string)
    return t
for l2 in sys.stdin.readlines():
        sys.stdout.write(relabel_newick(l2))''' % (str(revnamemap), tag)
        # _LOG.info("Use this code to map back internal nodes::\n %s" %
        #  str(self._revscript))

    def rename_script(self):
        return self._revscript

    def write_newick_node(self, node, out):
        child_nodes = node.child_nodes()
        if child_nodes:
            out.write('(')
            f_child = child_nodes[0]
            for child in child_nodes:
                if child is not f_child:
                    out.write(',')
                self.write_newick_node(child, out)
            out.write(')')

        if node.is_leaf() or not self._namemap:
            out.write("%s" % node._get_node_token())
        elif node in self._namemap:
            out.write("%s" % self._namemap[node])
        e = node.edge
        if e:
            sel = e.length
            if sel is not None:
                s = ""
                try:
                    s = float(sel)
                    s = str(s)
                except ValueError:
                    s = str(sel)
                if s:
                    out.write(":%s[%s]" % (s, e.label))

    def get_tree(self):
        return self._tree
    den_tree = property(get_tree)

    def count_leaves(self):
        return len(self._tree.leaf_nodes())

    def count_nodes(self):
        return len(self._tree.nodes())

    def calc_splits(self):
        for i in self._tree.postorder_edge_iter():
            nd = i.head_node
            if nd.is_leaf():
                i.num_leaves_below = 1
            else:
                i.num_leaves_below = sum([j.edge.num_leaves_below
                                          for j in nd.child_nodes()])

    def get_clade_edge(self, minSize):
        root = self._tree.seed_node
        # should only be 2 children, but in unlikely event that clade-based
        # decomp results in 1 child, then we should handle in some way
        root_children = root.child_nodes()
        assert len(root_children) == 2

        # Always break first child edge
        clade_edge = root.child_nodes()[0].edge
        assert clade_edge is not None
        return clade_edge

    def get_centroid_edge(self, minSize):
        """Get centroid edge"""
        root = self._tree.seed_node
        root_children = root.child_nodes()
        if root_children and (not hasattr(root_children[0].edge,
                                          "num_leaves_below")):
            self.calc_splits()
            n_leaves = self.count_leaves()
        else:
            if root.edge:
                n_leaves = root.edge.num_leaves_below
            else:
                n_leaves = sum([j.edge.num_leaves_below
                                for j in root_children])
        centroid_edge = None
        centroid_imbalance = n_leaves
        if (n_leaves <= minSize):
            return None
        half_taxa = n_leaves / 2
        for edge in self._tree.postorder_edge_iter():
            if edge.tail_node is None:
                continue
            n_descendants = edge.num_leaves_below
            if n_descendants > 1:
                pass
            if (
                (minSize is not None) and
                ((n_descendants < minSize) or
                 (self.n_leaves - n_descendants < minSize))
               ):
                continue
            imbalance = abs(half_taxa - n_descendants)
            if (imbalance < centroid_imbalance):
                centroid_edge = edge
                centroid_imbalance = imbalance
            assert centroid_edge is not None
        return centroid_edge

    def get_longest_edge(self, minSize):
        longest_edge = None
        longest_len = -1.0
        for edge in self._tree.postorder_edge_iter():
            if edge.tail_node is None:
                continue
            onesideSize = len(edge.head_node.leaf_nodes())
            if (
                (minSize is not None) and
                ((onesideSize < minSize) or
                 (self.n_leaves - onesideSize < minSize))
               ):
                continue
            if edge.length is not None and edge.length > longest_len:
                longest_edge = edge
                longest_len = edge.length
        assert longest_edge is not None
        return longest_edge

    def get_adjacent_edges(self, e):
        he = [i for i in e.head_node.get_incident_edges() if i is not e]
        te = [i for i in e.tail_node.get_incident_edges() if i is not e]
        he.extend(te)
        return he

    def get_breaking_edge(self, option, minSize):
        if option.lower() == 'centroid':
            return self.get_centroid_edge(minSize)
        elif option.lower() == 'longest':
            return self.get_longest_edge(minSize)
        elif option.lower() == 'clade':
            return self.get_clade_edge(minSize)
        else:
            raise ValueError('Unknown break strategy "%s"' % option)

    def bipartition_by_root(self):
        if (self.n_leaves == 1):
            return (None, None, None)
        root = self._tree.seed_node
        t1_root = root._child_nodes[0]
        t = self._tree
        t.prune_subtree(t1_root, update_splits=True, delete_outdegree_one=True)
        t1 = PhylogeneticTree(t)
        t2 = PhylogeneticTree(Tree(t1_root))
        # Reroot if there's more than node left
        if (t2.n_leaves > 1):
            t2._tree.reroot_at_node(t1_root)
        return t1, t2, root

    def bipartition_by_edge(self, e):
        """Prunes the subtree that attached to the head_node of edge e and
           returns them as a separate tree."""

        t = self._tree
        nr = e.head_node
        assert e.tail_node is not None
        assert e.head_node is not None
        assert nr.parent_node is e.tail_node
        is_valid_tree(t)

        potentially_deleted_nd = e.tail_node
        grandparent_nd = potentially_deleted_nd.parent_node
        e.tail_node.remove_child(nr, suppress_unifurcations=True)

        nr.edge.length = None
        nr.parent_node = None
        convert_node_to_root_polytomy(nr)
        t1 = PhylogeneticTree(Tree(seed_node=nr))
        # temp we could speed this up,
        # by telling the Phylogenetic tree how many leaves it has
        n1 = t1.n_leaves

        if hasattr(e, "num_leaves_below"):
            if grandparent_nd is None:
                old_root = potentially_deleted_nd
                if old_root.edge:
                    old_root.edge.num_leaves_below -= n1
            else:
                if potentially_deleted_nd in grandparent_nd.child_nodes():
                    potentially_deleted_nd.edge.num_leaves_below -= n1
                old_root = grandparent_nd
                if old_root.edge:
                    old_root.edge.num_leaves_below -= n1
                while old_root.parent_node:
                    old_root = old_root.parent_node
                    if old_root.edge:
                        old_root.edge.num_leaves_below -= n1
        else:
            old_root = grandparent_nd or potentially_deleted_nd
            while old_root.parent_node:
                old_root = old_root.parent_node

        t2 = PhylogeneticTree(Tree(seed_node=old_root))

        is_valid_tree(t1._tree)
        is_valid_tree(t2._tree)
        return t1, t2

    def leaf_node_names(self):
        leaves = self._tree.leaf_nodes()
        return [i.taxon.label for i in leaves]

    def compose_newick(self, labels=False):
        if not labels:
            return self._tree.as_string(
                schema="newick", suppress_rooting=True,
                suppress_internal_node_labels=not labels)
        else:
            stringIO = StringIO()
            self.write_newick_node(self._tree.seed_node, stringIO)
            ret = stringIO.getvalue()
            stringIO.close()
            return ret

    def write_newick_to_path(self, path):
        tree_handle = open(path, "w")
        tree_handle.write(self.compose_newick())
        tree_handle.write("")
        tree_handle.close()

    def read_tree_from_file(self, treefile, file_format):
        dataset = Dataset()
        dataset.read(open(treefile, 'rU'), schema=file_format)
        dendropy_tree = dataset.trees_blocks[0][0]
        self._tree = dendropy_tree
        self.n_leaves = self.count_leaves()

    def get_subtree(self, taxa):
        if len(taxa) == 0:
            return None
        tree = Tree(self._tree)
        if isinstance(taxa[0], str):
            tree.prune_taxa_with_labels(taxa)
        elif isinstance(taxa[0], Taxon):
            tree.prune_taxa(taxa)
        return PhylogeneticTree(tree)

    def bisect_tree(self, breaking_edge_style='centroid', minSize=None):
        """Partition 'tree' into two parts
        """
        snl = self.n_leaves
        if (breaking_edge_style != 'clade'):
            e = self.get_breaking_edge(breaking_edge_style, minSize)
            if (e is None):
                return None, None, None
            _LOG.debug("breaking_edge length = %s, %s" % (
                e.length, breaking_edge_style))
            tree1, tree2 = self.bipartition_by_edge(e)
        else:
            tree1, tree2, e = self.bipartition_by_root()

        _LOG.debug("Tree 1 has %s nodes, tree 2 has %s nodes" % (
            tree1.n_leaves, tree2.n_leaves))
        assert snl == tree1.n_leaves + tree2.n_leaves
        return tree1, tree2, e

#    def decompose_tree(self, maxSize, strategy, minSize = None, tree_map={},
#                       decomp_strategy='normal', pdistance=1, distances=None):
    def decompose_tree(self, maxSize, strategy, minSize=None, tree_map={},
                       decomp_strategy='normal', pdistance=1, distances=None,
                       maxDiam=None):
        """
        This function decomposes the tree until all subtrees are smaller than
        the max size, but does not decompose below min size.
        Two possible decompositions strategies can used: "centroid" and
        "longest".
        Returns a map containing the subtrees, in an ordered fashion.

        SIDE EFFECT: deroots the tree (TODO: necessary?)
        """
        # uym2 added #
        if (decomp_strategy in ["midpoint", "centroid"]):
            T = decompose_by_diameter(
                self._tree, strategy=decomp_strategy, max_size=maxSize,
                max_diam=maxDiam, min_size=minSize)
            for i, t in enumerate(T):
                tree_map[i] = PhylogeneticTree(t)
            return tree_map
        ##############

        # Don't deroot if doing clade-based decomposition
        if (strategy != 'clade'):
            self._tree.deroot()
        else:
            # If doing clade-based decomp and it's not rooted, root it!
            if self._tree.is_rooted is False:
                self._tree.reroot_at_midpoint()

        if (
            (decomp_strategy == 'hierarchical') and
            (self.count_leaves() > maxSize)
           ):
            tree_map[len(tree_map)] = copy.deepcopy(self)
        if (
            (self.count_leaves() > maxSize) or
            ((pdistance != 1) and
             (get_pdistance(distances, self.leaf_node_names()) > pdistance))
           ):
            (t1, t2, e) = self.bisect_tree(strategy, minSize)
            if e is not None:
                t1.decompose_tree(maxSize, strategy, minSize, tree_map,
                                  decomp_strategy, pdistance, distances)
                t2.decompose_tree(maxSize, strategy, minSize, tree_map,
                                  decomp_strategy, pdistance, distances)
            else:
                tree_map[len(tree_map)] = self
                _LOG.warning(
                    ("It was not possible to break-down the following tree "
                     "according to given subset sizes: %d , %d:\n %s") % (
                     minSize, maxSize, self._tree))
        else:
            tree_map[len(tree_map)] = self
        return tree_map

    def lable_edges(self):
        en = 0
        for e in self._tree.postorder_edge_iter():
            e.label = en
            en += 1

    '''
    Returns a given number of taxa that are closest to a given leaf
    '''
    def branchOut(self, centerTaxon, subsetSize, **kwargs):
        dist = {}
        pdm = treecalc.PatristicDistanceMatrix(self.den_tree)
        for i, s in enumerate(self.den_tree.taxon_set):  # @UnusedVariable
            if "filterTaxon" in kwargs:
                if not kwargs["filterTaxon"](s):
                    continue
            dist[s.label] = pdm(centerTaxon, s)
        incircle = sortByValue(dist)[0:subsetSize]
        return [node[0] for node in incircle]


def node_formatter(n):
    return str(id(n))


def edge_formatter(e):
    return "%s %f " % (str(id(e)), e.length)


def is_valid_tree(t):
    if (t.is_rooted):
        return True
    assert t and t
    rc = t.seed_node.child_nodes()
    num_children = len(rc)
    if num_children == 0:
        return True
    if num_children == 1:
        assert not rc[0].child_nodes()
        return True
    if num_children == 2:
        # What is with this code?  Why do we check the same variable twice?
        # Bug?  NN
        assert((not rc[0].child_nodes()) and (not rc[0].child_nodes()))
    return True



import dendropy

taxa = dendropy.TaxonNamespace()

t = dendropy.Tree.get_from_path(src='99_otus.tree', schema='newick',
                                taxon_namespace=taxa)
t2 = dendropy.Tree.get_from_path(
    src='sepp-package/ref/reference-gg-raxml-bl-rooted.tre', schema='newick',
    taxon_namespace=taxa)

t.encode_bipartitions()

missing = []
mapped = 0
for n in t.postorder_node_iter():
    if n.is_internal() and n.label is not None:
        # print (n.label)
        n.edge.bipartition.is_mutable = False
        if n.edge.bipartition not in t2.bipartition_edge_map:
            missing.append(n.edge.bipartition)
        else:
            t2n = t2.bipartition_edge_map[n.edge.bipartition].head_node
            t2n.label = n.label
            mapped = mapped + 1

a = t2.find_node_with_label(label="k__Archaea")

t2.reroot_at_node(a, update_bipartitions=False)

t2.write_to_path(dest='reference-gg-raxml-bl-rooted-relabelled.tre',
                 schema='newick', suppress_rooting=True,
                 suppress_internal_node_labels=False)



'''
Created on Oct 10, 2012

@author: smirarab
'''
from sepp.alignment import MutableAlignment, ExtendedAlignment
from sepp.exhaustive import JoinAlignJobs
from sepp.problem import SeppProblem
import glob

job_joiner = JoinAlignJobs
original_backbone_file = ('/projects/sate8/namphuon/ultra_large/1000000/'
                          'sate.fasta')
original_backbone = MutableAlignment()
done = original_backbone.read_filepath(original_backbone_file)

original_frag_file = ('/projects/sate8/namphuon/ultra_large/1000000/'
                      'initial.fas.100')
original_frag = MutableAlignment()
done = original_frag.read_filepath(original_frag_file)

# First build extended alignment on entire fragment set
extendedAlignment = ExtendedAlignment(original_frag.get_sequence_names())

dirs = glob.glob('/projects/sate8/namphuon/ultra_large/1000000/upp_100_10_new/'
                 'temp/upp.1_HNlM/root/P_0/A_0_*/')

dirs.reverse()
for dir in dirs:
    print("Working on %s\n" % dir)
    aligned_files = glob.glob('%sFC_*/hmmalign.results.*' % dir)
    sequence_files = glob.glob('%sFC_*/hmmalign.frag.*' % dir)
    base_alignment_file = glob.glob('%s/*.fasta' % dir)
    base_alignment = MutableAlignment()
    done = base_alignment.read_filepath(base_alignment_file[0])
    subbackbone = original_backbone.get_soft_sub_alignment(
        base_alignment.get_sequence_names())
    frags = MutableAlignment()
    sequence_names = []
    for file in sequence_files:
        seq = MutableAlignment()
        done = seq.read_filepath(file)
        done = sequence_names.extend(seq.get_sequence_names())
        for name, seq in seq.items():
            frags[name] = seq.upper()
    problem = SeppProblem(sequence_names)
    problem.set_subalignment(subbackbone)

    mut_subalg = problem.subalignment.get_mutable_alignment()
    remaining_cols = mut_subalg.delete_all_gap()
    problem.annotations["ref.alignment.columns"] = remaining_cols
    problem.fragments = frags
    ap_alg = problem.read_extendend_alignment_and_relabel_columns(
        base_alignment_file, aligned_files)
    extendedAlignment.merge_in(ap_alg, convert_to_string=False)

extendedAlignment.write_to_path("/projects/sate8/namphuon/ultra_large/1000000/"
                                "upp_100_10_new/upp.unmasked.fasta")
extendedAlignment.remove_insertion_columns()
extendedAlignment.write_to_path("/projects/sate8/namphuon/ultra_large/1000000/"
                                "upp_100_10_new/upp.masked.fasta")



'''
Created on Nov 14, 2012

@author: smirarab
'''


def gcd(a, b):
    """Return greatest common divisor using Euclid's Algorithm."""
    while b:
        a, b = b, a % b
    return a


def lcm(a, b):
    """Return lowest common multiple."""
    return a * b // gcd(a, b)



'''
Created on June 14, 2018

@author: Stefan.M.Janssen@gmail.com
'''
import sys
import tempfile
import shutil
import unittest
import sepp
from sepp.filemgr import get_data_path
import platform
from argparse import Namespace

from sepp.exhaustive import ExhaustiveAlgorithm
# sepp._DEBUG = True
# sepp.reset_loggers()


class Test(unittest.TestCase):
    x = None

    def setUp(self):
        # ensure necessary settings are made
        sepp.scheduler._jobPool = None
        sys.argv = [sys.argv[0], "-c", get_data_path("configs/test2.config")]
        self.x = ExhaustiveAlgorithm()
        self.x.options.alignment_file = open(
            get_data_path(
                "q2-fragment-insertion/reference_alignment_tiny.fasta"), "r")
        self.x.options.info_file = open(
            get_data_path(
                "q2-fragment-insertion/RAxML_info-reference-gg-raxml-bl.info"),
            "r")
        self.x.options.tree_file = open(
            get_data_path(
                "q2-fragment-insertion/reference_phylogeny_tiny.nwk"), "r")
        self.x.options.outdir = tempfile.mkdtemp()

        suff_bit = "-64" if sys.maxsize > 2**32 else "-32"
        if platform.system() == 'Darwin':
            suff_bit = ""
        for prog in ['hmmalign', 'hmmbuild', 'hmmsearch', 'pplacer']:
            setattr(self.x.options, prog, Namespace(
                path=get_data_path("../../../tools/bundled/%s/%s%s" % (
                    platform.system(), prog, suff_bit))))

    def tearDown(self):
        self.x.options.alignment_file.close()
        self.x.options.info_file.close()
        self.x.options.tree_file.close()
        self.x.options.fragment_file.close()
        sepp.scheduler._jobPool = None
        shutil.rmtree(self.x.options.outdir, ignore_errors=True)

    def test_id_collision_working(self):
        self.x.options.fragment_file = open(
            get_data_path(
                "q2-fragment-insertion/input_fragments.fasta"), "r")
        self.x.run()
        self.assertTrue(self.x.results is not None)

    def test_id_collision_collision(self):
        self.x.options.fragment_file = open(
            get_data_path(
                "q2-fragment-insertion/input_fragments_collide.fasta"), "r")
        with self.assertRaisesRegex(
                ValueError,
                ' whose names overlap with names in your reference'):
            self.x.run()
        self.assertTrue(self.x.results is None)

    def test_seqnames_whitespaces(self):
        self.x.options.fragment_file = open(
            get_data_path(
                "q2-fragment-insertion/input_fragments_spaces.fasta"), "r")
        with self.assertRaisesRegex(
                ValueError,
                "contain either whitespaces: "):
            self.x.run()
        self.assertTrue(self.x.results is None)

    def test_fake_jobs(self):
        self.x.options.fragment_file = open(
            get_data_path(
                "q2-fragment-insertion/input_fragments_tiny.fasta"), "r")
        self.x.run()
        self.assertTrue(self.x.results is not None)

    def test_notpiped_jobs(self):
        sepp.config.options().hmmsearch.piped = "False"
        self.x.options.fragment_file = open(
            get_data_path(
                "q2-fragment-insertion/input_fragments_tiny.fasta"), "r")
        self.x.run()
        self.assertTrue(self.x.results is not None)


if __name__ == "__main__":
    unittest.main()



'''
Created on Oct 10, 2012

@author: smirarab
'''
import sys
import random
import argparse
import os
import shutil
from sepp import get_logger
from sepp.alignment import MutableAlignment, ExtendedAlignment, _write_fasta
from sepp.exhaustive import JoinAlignJobs, ExhaustiveAlgorithm
from sepp.jobs import PastaAlignJob
from sepp.filemgr import get_temp_file
from sepp.config import options, valid_decomp_strategy
import sepp.config
from sepp.math_utils import lcm
from sepp.problem import SeppProblem

_LOG = get_logger(__name__)


class UPPJoinAlignJobs(JoinAlignJobs):
    '''
    After all alignments jobs for a placement subset have finished,
    we need to build those extended alignments. This join takes care of that
    step.
    '''
    def __init__(self):
        JoinAlignJobs.__init__(self)

    def perform(self):
        pp = self.placement_problem

        assert isinstance(pp, SeppProblem)
        pp.annotations["search_join_object"] = self

# Useful for multi-core merging if ever needed
# def mergetwo(x):
#    ((i,j),extended) = x
#    a=extended[i]
#    b=extended[j]
#    a.merge_in(b,convert_to_string=True)
#    extended[j] = None
#    extended[i] = a
#    del b
#    return "Success"


class UPPExhaustiveAlgorithm(ExhaustiveAlgorithm):
    '''
    This implements the exhaustive algorithm where all alignments subsets
    are searched for every fragment. This is for UPP, meaning that no placement
    is performed, and that there is always only one placement subset
    (currently).
    '''
    def __init__(self):
        ExhaustiveAlgorithm.__init__(self)
        self.pasta_only = False

    def generate_backbone(self):
        _LOG.info("Reading input sequences: %s" % (self.options.sequence_file))
        sequences = MutableAlignment()
        sequences.read_file_object(self.options.sequence_file)
        sequences.degap()
        fragments = MutableAlignment()
        if (options().median_full_length is not None):
            if (options().median_full_length == -1):
                seq_lengths = sorted(
                    [len(seq) for seq in list(sequences.values())])
                lengths = len(seq_lengths)
                l2 = int(lengths / 2)
                if lengths % 2:
                    options().median_full_length = (
                        seq_lengths[l2] + seq_lengths[l2 + 1]) / 2.0
                else:
                    options().median_full_length = seq_lengths[l2]

            (min_length, max_length) = (
                int(options().median_full_length * (
                    1 - options().backbone_threshold)),
                int(options().median_full_length*(
                    1 + options().backbone_threshold)))
            frag_names = [
                name for name in sequences
                if len(sequences[name]) > max_length or
                len(sequences[name]) < min_length]
            if (len(frag_names) > 0):
                _LOG.info(
                    "Detected %d fragmentary sequences" % len(frag_names))
                fragments = sequences.get_hard_sub_alignment(frag_names)
                [sequences.pop(i) for i in list(fragments.keys())]
        if (options().backbone_size is None):
            options().backbone_size = min(1000, int(sequences.get_num_taxa()))
            _LOG.info("Backbone size set to: %d" % (options().backbone_size))
        if (options().backbone_size > len(list(sequences.keys()))):
            options().backbone_size = len(list(sequences.keys()))
        sample = sorted(random.sample(
            sorted(list(sequences.keys())), options().backbone_size))
        backbone_sequences = sequences.get_hard_sub_alignment(sample)
        _LOG.debug("Backbone: %s" % (sorted(list(backbone_sequences.keys()))))
        [sequences.pop(i) for i in list(backbone_sequences.keys())]

        _LOG.info("Writing backbone set. ")
        backbone = get_temp_file("backbone", "backbone", ".fas")
        _write_fasta(backbone_sequences, backbone)

        _LOG.info("Generating pasta backbone alignment and tree. ")
        pastaalignJob = PastaAlignJob()
        moleculeType = options().molecule
        if (options().molecule == 'amino'):
            moleculeType = 'protein'
        pastaalignJob.setup(backbone, options().backbone_size,
                            moleculeType, options().cpu)
        pastaalignJob.run()
        (a_file, t_file) = pastaalignJob.read_results()

        shutil.copyfile(t_file, self.get_output_filename("pasta.fasttree"))
        shutil.copyfile(a_file, self.get_output_filename("pasta.fasta"))

        options().placement_size = self.options.backbone_size
        options().alignment_file = open(
            self.get_output_filename("pasta.fasta"))
        options().tree_file = open(self.get_output_filename("pasta.fasttree"))
        _LOG.info(
            "Backbone alignment written to %s.\nBackbone tree written to %s"
            % (options().alignment_file, options().tree_file))
        sequences.set_alignment(fragments)
        if (len(sequences) == 0):
            sequences = MutableAlignment()
            sequences.read_file_object(open(self.options.alignment_file.name))
            self.results = ExtendedAlignment(fragment_names=[])
            self.results.set_alignment(sequences)
            _LOG.info(
                "No query sequences to align.  Final alignment saved as %s"
                % self.get_output_filename("alignment.fasta"))
            self.output_results()
            sys.exit(0)
        else:
            query = get_temp_file("query", "backbone", ".fas")
            options().fragment_file = query
            _write_fasta(sequences, query)

    def check_options(self):
        self.check_outputprefix()
        options().info_file = "A_dummy_value"

        # Check to see if tree/alignment/fragment file provided, if not,
        # generate it from sequence file
        if (
            (not options().tree_file is None) and
            (not options().alignment_file is None) and
            (not options().sequence_file is None)
           ):
            options().fragment_file = options().sequence_file
        elif (
              (options().tree_file is None) and
              (options().alignment_file is None) and
              (not options().sequence_file is None)
             ):
            self.generate_backbone()
        else:
            _LOG.error(
                ("Either specify the backbone alignment and tree and query "
                 "sequences or only the query sequences.  Any other "
                 "combination is invalid"))
            exit(-1)
        sequences = MutableAlignment()
        sequences.read_file_object(open(self.options.alignment_file.name))
        backbone_size = sequences.get_num_taxa()
        if options().backbone_size is None:
            options().backbone_size = backbone_size
        assert options().backbone_size == backbone_size, (
            ("Backbone parameter needs to match actual size of backbone; "
             "backbone parameter:%s backbone_size:%s")
            % (options().backbone_size, backbone_size))
        if options().placement_size is None:
            options().placement_size = options().backbone_size
        return ExhaustiveAlgorithm.check_options(self)

    def merge_results(self):
        assert \
            len(self.root_problem.get_children()) == 1, \
            "Currently UPP works with only one placement subset."
        '''
        Merge alignment subset extended alignments to get one extended
        alignment for current placement subset.
        '''
        pp = self.root_problem.get_children()[0]
        _LOG.info(
            "Merging sub-alignments for placement problem : %s." % (pp.label))
        ''' First assign fragments to the placement problem'''
        pp.fragments = pp.parent.fragments.get_soft_sub_alignment([])
        for ap in pp.get_children():
            pp.fragments.seq_names |= set(ap.fragments)

        ''' Then Build an extended alignment by merging all hmmalign results'''
        _LOG.debug(
            "fragments are %d:\n %s" % (
                len(pp.fragments.seq_names), pp.fragments.seq_names))
        extendedAlignment = ExtendedAlignment(pp.fragments.seq_names)
        for ap in pp.children:
            assert isinstance(ap, SeppProblem)
            ''' Get all fragment chunk alignments for this alignment subset'''
            aligned_files = [fp.get_job_result_by_name('hmmalign') for
                             fp in ap.children if
                             fp.get_job_result_by_name('hmmalign') is not None]
            _LOG.debug(
                "Merging fragment chunks for subalignment : %s." % (ap.label))
            ap_alg = ap.read_extendend_alignment_and_relabel_columns(
                ap.jobs["hmmbuild"].infile, aligned_files)
            _LOG.debug(
                "Merging alignment subset into placement subset: %s." %
                (ap.label))
            extendedAlignment.merge_in(ap_alg, convert_to_string=False)

        extendedAlignment.from_bytearray_to_string()
        self.results = extendedAlignment

# Useful for multi-core merging if ever needed
#    def parallel_merge_results(self):
#        assert len(self.root_problem.get_children()) == 1, "Currently UPP
#        works with only one placement subset."
#        '''
#        Merge alignment subset extended alignments to get one extended
#        alignment
#        for current placement subset.
#        '''
#        pp = self.root_problem.get_children()[0]
#        _LOG.info("Merging sub-alignments for placement problem : %s."
#        %(pp.label))
#        ''' Then Build an extended alignment by merging all hmmalign
#            results'''
#        manager = Manager()
#        extendedAlignments = manager.list()
#        for ap in pp.children:
#            assert isinstance(ap, SeppProblem)
#            ''' Get all fragment chunk alignments for this alignment subset'''
#            aligned_files = [fp.get_job_result_by_name('hmmalign') for
#                             fp in ap.children if
#                             fp.get_job_result_by_name('hmmalign')
#                               is not None]
#            _LOG.info("Merging fragment chunks for subalignment : %s."
#                       %(ap.label))
#            ap_alg = ap.read_extendend_alignment_and_relabel_columns\
#                        (ap.jobs["hmmbuild"].infile , aligned_files)
#            _LOG.info("Merging alignment subset into placement subset: %s."
#                          %(ap.label))
#            extendedAlignments.append(ap_alg)
#
#        while len(extendedAlignments)>1:
#            a=range(0,len(extendedAlignments))
#            #print [len(x) for x in extendedAlignments]
#            x = zip(a[0::2],a[1::2])
#            mapin = zip (x,[extendedAlignments]*len(x))
#            _LOG.debug("One round of merging started. Currently have %d
#                        alignments left. " %len(extendedAlignments))
#            Pool(max(12,len(extendedAlignments))).map(mergetwo,mapin)
#            #print [len(x) if x is not None else "None" for x in
#                    extendedAlignments]
#            extendedAlignments = manager.list([x for x in
#                             extendedAlignments if x is not None])
#            extendedAlignments.reverse()
#            _LOG.debug("One round of merging finished. Still have %d
#                       alignments left. " %len(extendedAlignments))
#        extendedAlignment = extendedAlignments[0]
#        extendedAlignment.from_bytearray_to_string()
#        self.results = extendedAlignment

    def output_results(self):
        extended_alignment = self.results
        _LOG.info("Generating output. ")
        outfilename = self.get_output_filename("alignment.fasta")
        extended_alignment.write_to_path(outfilename)
        _LOG.info("Unmasked alignment written to %s" % outfilename)
        outfilename = self.get_output_filename("insertion_columns.txt")
        extended_alignment.write_insertion_column_indexes(outfilename)
        _LOG.info("The index of insertion columns written to %s" % outfilename)
        extended_alignment.remove_insertion_columns()
        outfilename = self.get_output_filename("alignment_masked.fasta")
        extended_alignment.write_to_path(outfilename)
        _LOG.info("Masked alignment written to %s" % outfilename)

    def check_and_set_sizes(self, total):
        assert (self.options.placement_size is None) or (
                self.options.placement_size >= total), \
                ("currently UPP works with only one placement subset."
                 " Please leave placement subset size option blank.")
        ExhaustiveAlgorithm.check_and_set_sizes(self, total)
        self.options.placement_size = total

    def _get_new_Join_Align_Job(self):
        return UPPJoinAlignJobs()

    def modify_tree(self, a_tree):
        ''' Filter out taxa on long branches '''
        self.filtered_taxa = []
        if self.options.long_branch_filter is not None:
            tr = a_tree.get_tree()
            elen = {}
            for e in tr.leaf_edge_iter():
                elen[e] = e.length
            elensort = sorted(elen.values())
            mid = elensort[len(elensort) / 2]
            torem = []
            for k, v in list(elen.items()):
                if v > mid * self.options.long_branch_filter:
                    self.filtered_taxa.append(k.head_node.taxon.label)
                    torem.append(k.head_node.taxon)
            tr.prune_taxa(torem)

    def create_fragment_files(self):
        alg_subset_count = len(list(self.root_problem.iter_leaves()))
        frag_chunk_count = lcm(
            alg_subset_count, self.options.cpu) // alg_subset_count
        _LOG.info(
            "%d taxa pruned from backbone and added to fragments: %s"
            % (len(self.filtered_taxa), " , ".join(self.filtered_taxa)))
        return self.read_and_divide_fragments(
            frag_chunk_count,
            extra_frags=self.root_problem.subalignment.get_soft_sub_alignment(
                self.filtered_taxa))


def augment_parser():
    root_p = open(os.path.join(os.path.split(
        os.path.split(__file__)[0])[0], "home.path")).readlines()[0].strip()
    upp_config_path = os.path.join(root_p, "upp.config")
    sepp.config.set_main_config_path(upp_config_path)
    parser = sepp.config.get_parser()
    parser.description = (
        "This script runs the UPP algorithm on set of sequences.  A backbone"
        " alignment and tree can be given as input.  If none is provided, a"
        " backbone will be automatically generated.")

    decompGroup = parser.groups['decompGroup']
    decompGroup.__dict__['description'] = ' '.join(
        ["These options",
         ("determine the alignment decomposition size, backbone size,"
          " and how to decompose the backbone set.")])

    decompGroup.add_argument(
        "-A", "--alignmentSize", type=int,
        dest="alignment_size", metavar="N",
        default=10,
        help="max alignment subset size of N "
             "[default: 10]")
    decompGroup.add_argument(
        "-M", "--median_full_length", type=int,
        dest="median_full_length", metavar="N",
        default=None,
        help="Consider all fragments that are 25%% longer or shorter than N "
             "to be excluded from the backbone.  If value is -1, then UPP will"
             " use the median of the sequences as the median full length "
             "[default: None]")
    decompGroup.add_argument(
        "-T", "--backbone_threshold", type=float,
        dest="backbone_threshold", metavar="N",
        default=0.25,
        help="Only consider sequences with lengths within (1-N)*M and (1+N)*"
             "M as full-length, where M is the median length of the full-"
             "length sequence given by the -M option and N is the percentage"
             " given by the -T option."
             "[default: 0.25]")
    decompGroup.add_argument(
        "-B", "--backboneSize", type=int,
        dest="backbone_size", metavar="N",
        default=None,
        help="(Optional) size of backbone set.  "
             "If no backbone tree and alignment is given, the sequence file "
             "will be randomly split into a backbone set (size set to N) and "
             "query set (remaining sequences), [default: min(1000,input "
             "size)]")
    decompGroup.add_argument(
        "-S", "--decomp_strategy", type=valid_decomp_strategy,
        dest="decomp_strategy", metavar="DECOMP",
        default="hierarchical",
        help="decomposition strategy "
             "[default: ensemble of HMMs (hierarchical)]")

    inputGroup = parser.groups['inputGroup']
    inputGroup .add_argument(
        "-s", "--sequence_file", type=argparse.FileType('r'),
        dest="sequence_file", metavar="SEQ",
        default=None,
        help="Unaligned sequence file.  "
             "If no backbone tree and alignment is given, the sequence file "
             "will be randomly split into a backbone set (size set to B) and "
             "query set (remaining sequences), [default: None]")
    inputGroup.add_argument(
        "-c", "--config",
        dest="config_file", metavar="CONFIG",
        type=argparse.FileType('r'),
        help="A config file, including options used to run UPP. Options "
             "provided as command line arguments overwrite config file values"
             " for those options. "
             "[default: %(default)s]")
    inputGroup.add_argument(
        "-t", "--tree",
        dest="tree_file", metavar="TREE",
        type=argparse.FileType('r'),
        help="Input tree file (newick format) "
             "[default: %(default)s]")
    inputGroup.add_argument(
        "-a", "--alignment",
        dest="alignment_file", metavar="ALIGN",
        type=argparse.FileType('r'),
        help="Aligned fasta file "
             "[default: %(default)s]")

    uppGroup = parser.add_argument_group(
        "UPP Options".upper(),
        "These options set settings specific to UPP")

    uppGroup.add_argument(
        "-l", "--longbranchfilter", type=int,
        dest="long_branch_filter", metavar="N",
        default=None,
        help="Branches longer than N times the median branch length are "
             "filtered from backbone and added to fragments."
             " [default: None (no filtering)]")

    seppGroup = parser.add_argument_group(
        "SEPP Options".upper(),
        ("These options set settings specific to SEPP and are not used"
         " for UPP."))
    seppGroup.add_argument(
        "-P", "--placementSize", type=int,
        dest="placement_size", metavar="N",
        default=None,
        help="max placement subset size of N "
             "[default: 10%% of the total number of taxa]")
    seppGroup.add_argument(
        "-r", "--raxml",
        dest="info_file", metavar="RAXML",
        type=argparse.FileType('r'),
        help="RAxML_info file including model parameters, generated by RAxML."
             "[default: %(default)s]")
    seppGroup.add_argument(
        "-f", "--fragment",
        dest="fragment_file", metavar="FRAG",
        type=argparse.FileType('r'),
        help="fragment file "
             "[default: %(default)s]")


def main():
    augment_parser()
    UPPExhaustiveAlgorithm().run()


if __name__ == '__main__':
    main()



'''
Created on May 31, 2015

@author: namphuon
'''
import argparse
import os
from sepp import get_logger
from sepp.alignment import MutableAlignment
from sepp.exhaustive import ExhaustiveAlgorithm
from sepp.config import options, valid_decomp_strategy
import sepp.config
from sepp.math_utils import lcm
from sepp.problem import SeppProblem
from sepp.scheduler import JobPool, Join

_LOG = get_logger(__name__)


class EnsembleJoinSearchJobs(Join):
    '''
    After all search jobs have finished on tips, we need return the
    distribution of the bitscores for the search.  This join accomplishes this
    '''
    def __init__(self):
        Join.__init__(self)

    def setup_with_root_problem(self, root_problem):
        self.root_problem = root_problem
        for p in root_problem.iter_leaves():
            self.add_job(p.jobs["hmmsearch"])

    def perform(self):
        '''
        A dummy join that waits for all the search results to complete
        '''
        print("")

    def __str__(self):
        return "join search jobs for all tips of ", self.root_problem


class EnsembleExhaustiveAlgorithm(ExhaustiveAlgorithm):
    '''
    This implements the exhaustive algorithm where all alignments subsets
    are searched for every fragment. This is for UPP, meaning that no placement
    is performed, and that there is always only one placement subset
    (currently).
    '''
    def __init__(self):
        ExhaustiveAlgorithm.__init__(self)
        self.symfrac = False
        self.elim = None
        self.filters = True

    def check_options(self):
        options().info_file = "A_dummy_value"

        if options().tree_file is None or options().alignment_file is None:
            _LOG.error(
                "Specify the backbone alignment and tree and query sequences")
            exit(-1)
        sequences = MutableAlignment()
        sequences.read_file_object(open(self.options.alignment_file.name))
        return ExhaustiveAlgorithm.check_options(self)

    def check_and_set_sizes(self, total):
        assert (self.options.placement_size is None) or (
                self.options.placement_size >= total), \
                ("currently eHMMs works with only one placement subset. "
                 "Please leave placement subset size option blank.")
        ExhaustiveAlgorithm.check_and_set_sizes(self, total)
        self.options.placement_size = total

    def merge_results(self):
        ''' merges search results'''
        if "fragments.distribution.done" in self.root_problem.annotations:
            return
        sequence_scores = dict([(name, [])
                                for name
                                in self.root_problem.fragments.keys()])
        for fragment_chunk_problem in self.root_problem.iter_leaves():
            align_problem = fragment_chunk_problem.get_parent()
            assert isinstance(align_problem, SeppProblem)
            '''For each subproblem start with an empty set of fragments,
               and add to them as we encounter new best hits for that
               subproblem'''
            if align_problem.fragments is None:
                align_problem.fragments = \
                    self.root_problem.fragments.get_soft_sub_alignment([])
            search_res = fragment_chunk_problem.get_job_result_by_name(
                "hmmsearch")
            for key, val in search_res.items():
                sequence_scores[key].append([val[1], val[0]])

        # TODO: is the following efficient enough? Do we need to make lists
        # and then turn them to sets?
        notScored = []
        for key, v in sequence_scores.items():
            if len(v) == 0:
                notScored.append(key)

        self.root_problem.annotations["fragments.distribution.done"] = 1

        ''' Make sure all fragments are in at least one subproblem.
        TODO: what to do with those that are not?  For now, only output
        warning message'''
        # notScored = [k for k, v in max_evalues.iteritems() if v[1] is None]
        _LOG.warning("Fragments %s are not scored against any subset" %
                     str(notScored))
        # assert len(notScored) == 0,
        # "Fragments %s are not scored against any subset" %str(notScored)
        self.results = sequence_scores

    def connect_jobs(self):
        ''' a callback function called after hmmbuild jobs are finished'''
        def enq_job_searchfragment(result, search_job):
            search_job.hmmmodel = result
            JobPool().enqueue_job(search_job)
        assert isinstance(self.root_problem, SeppProblem)
        for placement_problem in self.root_problem.get_children():
            '''For each alignment subproblem, ...'''
            for alg_problem in placement_problem.children:
                assert isinstance(alg_problem, SeppProblem)
                ''' create the build model job'''
                bj = alg_problem.jobs["hmmbuild"]
                ''' create the search jobs'''
                for fc_problem in alg_problem.get_children():
                    sj = fc_problem.jobs["hmmsearch"]
                    ''' connect bulid and search jobs'''
                    bj.add_call_Back(
                        lambda result, next_job=sj: enq_job_searchfragment(
                            result, next_job))
        jsj = EnsembleJoinSearchJobs()
        jsj.setup_with_root_problem(self.root_problem)

    def output_results(self):
        search_results = self.results
        _LOG.info("Generating csv of search results. ")
        outfilename = self.get_output_filename("scores.csv")
        not_matched = self.get_output_filename("unmatched.csv")
        f = open(outfilename, 'w')
        unmatched = open(not_matched, 'w')
        f.write("seq,bitscore,evalue\n")

        for key, value in search_results.items():
            if len(value) == 0:
                unmatched.write("%s " % key)
            else:
                for pair in value:
                    f.write("%s,%0.4f,%s\n" % (
                        key, pair[0], "{:.3e}".format(pair[1])))
        f.close()
        unmatched.close()

    def create_fragment_files(self):
        alg_subset_count = len(list(self.root_problem.iter_leaves()))
        frag_chunk_count = lcm(
            alg_subset_count, self.options.cpu) // alg_subset_count
        return self.read_and_divide_fragments(frag_chunk_count)


def augment_parser():
    sepp.config.set_main_config_path(os.path.expanduser("~/.sepp/upp.config"))
    parser = sepp.config.get_parser()
    parser.description = (
        "This script runs the UPP algorithm on set of sequences.  A backbone "
        "alignment and tree can be given as input.  If none is provided, a "
        "backbone will be automatically generated.")

    decompGroup = parser.groups['decompGroup']
    decompGroup.__dict__['description'] = ' '.join([
        "These options",
        ("determine the alignment decomposition size, backbone size, and how "
         "to decompose the backbone set.")])

    decompGroup.add_argument(
        "-A", "--alignmentSize", type=int,
        dest="alignment_size", metavar="N",
        default=10,
        help="max alignment subset size of N [default: 10]")
    decompGroup.add_argument(
        "-S", "--decomp_strategy", type=valid_decomp_strategy,
        dest="decomp_strategy", metavar="DECOMP",
        default="hierarchical",
        help="decomposition strategy "
             "[default: ensemble of HMMs (hierarchical)]")

    inputGroup = parser.groups['inputGroup']
    inputGroup .add_argument(
        "-s", "--sequence_file", type=argparse.FileType('r'),
        dest="sequence_file", metavar="SEQ",
        default=None,
        help="Unaligned sequence file.  "
             "If no backbone tree and alignment is given, the sequence file "
             "will be randomly split into a backbone set (size set to B) and "
             "query set (remaining sequences), [default: None]")
    inputGroup.add_argument(
        "-c", "--config",
        dest="config_file", metavar="CONFIG",
        type=argparse.FileType('r'),
        help="A config file, including options used to run UPP. Options "
             "provided as command line arguments overwrite config file values"
             " for those options. "
             "[default: %(default)s]")
    inputGroup.add_argument(
        "-t", "--tree",
        dest="tree_file", metavar="TREE",
        type=argparse.FileType('r'),
        help="Input tree file (newick format) "
             "[default: %(default)s]")
    inputGroup.add_argument(
        "-a", "--alignment",
        dest="alignment_file", metavar="ALIGN",
        type=argparse.FileType('r'),
        help="Aligned fasta file "
             "[default: %(default)s]")

    parser.add_argument_group(
        "UPP Options".upper(),
        "These options set settings specific to UPP")

    seppGroup = parser.add_argument_group(
        "SEPP Options".upper(),
        "These options set settings specific to SEPP and are not used for "
        "UPP.")
    seppGroup.add_argument(
        "-P", "--placementSize", type=int,
        dest="placement_size", metavar="N",
        default=None,
        help="max placement subset size of N "
             "[default: 10%% of the total number of taxa]")
    seppGroup.add_argument(
        "-r", "--raxml",
        dest="info_file", metavar="RAXML",
        type=argparse.FileType('r'),
        help="RAxML_info file including model parameters, generated by RAxML."
             "[default: %(default)s]")
    seppGroup.add_argument(
        "-f", "--fragment",
        dest="fragment_file", metavar="FRAG",
        type=argparse.FileType('r'),
        help="fragment file [default: %(default)s]")


def main():
    augment_parser()
    EnsembleExhaustiveAlgorithm().run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python

###########################################################################
#    Copyright 2012 Siavash Mirarab, Nam Nguyen, and Tandy Warnow.
#    This file is part of SEPP.
#
#    SEPP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SEPP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SEPP.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################


from sepp import get_logger, is_temp_kept
import tempfile
import re
import sepp

"""
File and path management.
"""

import os
import shutil
import inspect

_LOG = get_logger(__name__)


def open_with_intermediates(filepath, mode):
    """Opens a `filepath` in the `mode`, but will create intermediate
       directories if they are absent."""
    d = os.path.dirname(filepath)
    if d:
        if not os.path.exists(d):
            os.makedirs(d)
        elif not os.path.isdir(d):
            raise IOError(
                ('The file "%s" cannot be created because "%s" exists but is '
                 'not a directory') % (filepath, d))
    return open(filepath, mode)


def get_temp_file(prefix, relative_path, suffix=""):
    '''
    Returns a temporary directory given a relative_path (relative to root temp
    directory) and a suffix. Use tempfile_for_subproblem when possible.
    '''
    d = os.path.join(get_root_temp_dir(), relative_path)
    if not os.path.exists(d):
        os.makedirs(d)
    return tempfile.mktemp(prefix=prefix,
                           suffix=suffix,
                           dir=d)


def tempdir_for_subproblem(subproblem):
    '''
    Returns the path to a temporary directory for a given subproblem.
    '''
    largs = [x.label for x in reversed(subproblem.get_path_to_root())]
    return os.path.join(*largs)


def tempfile_for_subproblem(prefix, subproblem, suffix=""):
    '''
    Returns the path to a temporary file for a given subproblem, and a prefix.
    '''
    from sepp.problem import SeppProblem
    assert isinstance(subproblem, SeppProblem)
    return get_temp_file(prefix, tempdir_for_subproblem(subproblem), suffix)


def is_my_temp_file(path):
    '''
    is this file part of my temp directory structure?
    '''
    return os.path.commonprefix(
        [get_root_temp_dir(), path]) == get_root_temp_dir()


def remove_temp(path):
    '''
    Tries to remove a temp directory. Does so only if a temp file is created
    inside the tmp directory structure of current run.
    '''
    if not is_my_temp_file(path):
        _LOG.warning(
            ("Temp File %s is not under temporary structure created by"
             " this job. It will not be removed. "))
        return
    if not is_temp_kept() and os.path.exists(path):
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)


def get_default_temp_dir():
    ''' Returns a default temp directory. This is used internally, in case
    a temp directory is not provided by the user. Use get_root_temp_dir to find
    the current root temp directory.
    '''
    tempfile.gettempdir()
    return os.path.join(tempfile.tempdir, "sepp")


_root_temp_dir = None


def get_root_temp_dir():
    '''
    Return the root temp directory for the current run. Creates the temp file
    the first time it is called in any run.
    '''
    global _root_temp_dir
    if _root_temp_dir is None:
        _root_temp_dir = tempfile.mkdtemp(
            prefix=sepp.config.options().output + ".",
            dir=sepp.config.options().tempdir)
        _LOG.info("Root temp directory built: %s" % _root_temp_dir)
    return _root_temp_dir


def set_root_temp_dir(tmp):
    global _root_temp_dir
    _root_temp_dir = tmp


def check_or_make_dir_path(path):
    '''
    Make sure a path to a directory exists, and return the normalized path.
    If the path is to a file, or if it cannot be made, return None.
    '''
    if not os.path.exists(path):
        # Race condition exists here
        try:
            os.mkdir(path)
        except FileExistsError:
            return os.path.normpath(os.path.abspath(path))
    if os.path.exists(path) and os.path.isdir(path):
        return os.path.normpath(os.path.abspath(path))
    return None


def directory_has_files_with_prefix(d, prefix):
    '''
    Check to see if a directory has files with a given prefix
    '''
    files = os.listdir(d)
    for f in files:
        if (re.search('^' + prefix + '.*', f) is not None):
            return True
    return False


# copy and paste from https://github.com/biocore/scikit-bio/blob/
# aa502f9dbc7670cf1f8feb6777f6abf181deadc5/skbio/util/_testing.py#L49
def get_data_path(fn, subfolder='data'):
    """Return path to filename ``fn`` in the data folder.
    During testing it is often necessary to load data files. This
    function returns the full path to files in the ``data`` subfolder
    by default.
    Parameters
    ----------
    fn : str
        File name.
    subfolder : str, defaults to ``data``
        Name of the subfolder that contains the data.
    Returns
    -------
    str
        Inferred absolute path to the test data for the module where
        ``get_data_path(fn)`` is called.
    Notes
    -----
    The requested path may not point to an existing file, as its
    existence is not checked.
    """
    # getouterframes returns a list of tuples: the second tuple
    # contains info about the caller, and the second element is its
    # filename
    callers_filename = inspect.getouterframes(inspect.currentframe())[1][1]
    path = os.path.dirname(os.path.abspath(callers_filename))
    data_path = os.path.join(path, subfolder, fn)

    return data_path



#!/usr/bin/env python

# ##########################################################################
#    Copyright 2012 Siavash Mirarab, Nam Nguyen, and Tandy Warnow.
#    This file is part of SEPP.
#
#    SEPP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SEPP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SEPP.  If not, see <http://www.gnu.org/licenses/>.
# ##########################################################################

import os
import platform
import sys

# from distutils.core import setup
from distribute_setup import use_setuptools
import shutil
from setuptools import find_packages
from distutils.core import setup, Command
from distutils.spawn import find_executable

use_setuptools(version="0.6.24")
version = "4.3.10"


def get_tools_dir(where):
    platform_name = platform.system()
    if where is None:
        where = os.path.join("bundled", platform_name)
    path = os.path.join(os.getcwd(), "tools", where)
    if not os.path.exists(path):
        raise OSError("SEPP does not bundle tools for '%s' at this time!" %
                      platform_name)
    return path


def get_tool_name(tool, bits):
    if platform.system() == "Darwin" or not bits:  # MAC doesn't have 32/64
        return tool
    is_64bits = sys.maxsize > 2**32
    return "%s-%s" % (tool, "64" if is_64bits else "32")


class ConfigSepp(Command):
    """setuptools Command"""
    description = "Configures Sepp for the current user"
    user_options = [(
        'contained', 'c',
        ("Whether SEPP should be installed in a self-contained "
         "manner or on user's home"))]

    def initopts(self):
        self.contained = None
        self.configfile = None
        self.basepath = None

    def initpath(self, name):
        if self.contained:
            self.configfile = os.path.expanduser(
                os.path.abspath(os.path.join(".sepp", name)))
            self.basepath = os.path.dirname(self.configfile)
        else:
            self.configfile = os.path.expanduser("~/.sepp/%s" % name)
            self.basepath = os.path.expanduser("~/.sepp")
        with open('home.path', 'w') as fo:
            fo.write(self.basepath)
            fo.close()

    def get_tools_dest(self):
        return os.path.join(self.basepath, "bundled-v%s" % version)

    def copy_tool_to_lib(self, tool, where=None, bits=True):
        shutil.copy2(
            os.path.join(get_tools_dir(where), get_tool_name(tool, bits)),
            os.path.join(self.get_tools_dest(), tool))

    def initialize_options(self):
        """init options"""
        self.initopts()

    def finalize_options(self):
        """finalize options"""
        self.initpath("main.config")
        print("\nCreating main sepp config file at %s and tools at %s" % (
            self.configfile, self.basepath))

    def run(self):
        def get_tool_name(tool, bits):
            # MAC doesn't have 32/64
            if platform.system() == "Darwin" or not bits:
                return tool
            is_64bits = sys.maxsize > 2**32
            return "%s-%s" % (tool, "64" if is_64bits else "32")

        # Create the default config file
        if not os.path.exists(self.basepath):
            os.mkdir(self.basepath)
        if not os.path.exists(self.get_tools_dest()):
            os.mkdir(self.get_tools_dest())
        c = open("default.main.config")
        d = open(self.configfile, "w")
        for l1 in c:
            l1 = l1.replace("~", self.get_tools_dest())
            d.write(l1)
        d.close()

        # Copy tools to a bundled directory inside .sepp
        self.copy_tool_to_lib("guppy")
        self.copy_tool_to_lib("pplacer")
        self.copy_tool_to_lib("hmmalign")
        self.copy_tool_to_lib("hmmsearch")
        self.copy_tool_to_lib("hmmbuild")
        # TODO: should we compile and build merge.jar?
        self.copy_tool_to_lib("seppJsonMerger.jar", where="merge", bits=False)


class ConfigUPP(ConfigSepp):
    """setuptools Command"""
    description = "Configures UPP for the current user"
    user_options = [('contained', 'c',
                     ("Whether SEPP should be installed in a self-contained "
                      "manner or on user's home"))]

    def initialize_options(self):
        """init options"""
        self.initopts()

    def finalize_options(self):
        """finalize options"""
        self.initpath("upp.config")
        print("\nCreating main UPP config file at %s and tools at %s" % (
            self.configfile, self.basepath))

    def run(self):
        # Create the default config file
        c = open("default.main.config")
        d = open(self.configfile, "w")
        for l2 in c:
            l2 = l2.replace("~", self.get_tools_dest())
            d.write(l2)
        d.write('\n[pasta]\npath=run_pasta.py\n')
        d.close()


class ConfigTIPP(ConfigSepp):
    """setuptools Command"""
    description = "Configures TIPP for the current user"
    user_options = [('contained', 'c',
                     ("Whether SEPP should be installed in a self-contained "
                      "manner or on user's home"))]

    def initialize_options(self):
        """init options"""
        self.initopts()

    def finalize_options(self):
        """finalize options"""
        self.initpath("tipp.config")
        print("\nCreating main TIPP config file at %s and tools at %s" % (
            self.configfile, self.basepath))

    def run(self):
        # Create the default config file
        c = open("default.main.config")
        d = open(self.configfile, "w")
        for l3 in c:
            l3 = l3.replace("~", self.get_tools_dest())
            if (l3.find('seppJsonMerger.jar') != -1):
                l3 = l3.replace('seppJsonMerger.jar', 'tippJsonMerger.jar')
            d.write(l3)
        if not os.getenv('SATE') is None:
            d.write('\n[sate]\npath=%s' % os.getenv('SATE'))
        if os.getenv('BLAST') is None and not find_executable("blastn"):
            print("\nWarning! BLAST variable is not defined.  If you plan to "
                  "run TIPP for abundance profiling,"
                  " then have BLAST pointed to blastn executable. "
                  "You can also change your config to point to"
                  " blastn by including the following line in your"
                  " config:\n[blast]\npath=/location/of/blast_directory/"
                  "blastn\n")
            d.write('\n[blast]\npath=None\n')
        else:
            blastn_path = find_executable("blastn") if os.getenv('BLAST') \
                is None else os.getenv('BLAST')
            d.write('\n[blast]\npath=%s\n' % blastn_path)

        if os.getenv('REFERENCE') is None:
            print("\nWarning! REFERENCE variable is not defined. "
                  "If you plan to run TIPP for abundance profiling, then have "
                  "REFERENCE pointed to Reference directory.  You can also "
                  "change your config to point to the Reference directory by "
                  "including the following line in your config:\n[reference]"
                  "\npath=/location/of/reference_directory/\n")
        d.write('\n[reference]\npath=%s\n' % os.getenv('REFERENCE'))
        d.write('\n[tipp]\npushdown = true\n')
        d.close()

        # Copy tools to a bundled directory inside .sepp
        self.copy_tool_to_lib("tippJsonMerger.jar", where="merge", bits=False)


setup(name="sepp",
      version=version,
      description="SATe enabled phylogenetic placement.",
      packages=find_packages(),

      url="https://github.com/smirarab/sepp",
      author="Siavash Mirarab and Nam Nguyen",
      author_email="smirarab@gmail.com, namphuon@cs.utah.edu",

      license="General Public License (GPL)",
      install_requires=["dendropy >= 4.0.0"],
      provides=["sepp"],
      scripts=["run_sepp.py", 'run_tipp.py', 'run_upp.py', 'run_abundance.py',
               "split_sequences.py", "run_tipp_tool.py"],
      cmdclass={"config": ConfigSepp, "tipp": ConfigTIPP, "upp": ConfigUPP},
      data_files=[('', ['home.path'])],

      classifiers=["Environment :: Console",
                   "Intended Audience :: Developers",
                   "Intended Audience :: Science/Research",
                   ("License :: OSI Approved :: GNU General Public "
                    "License (GPL)"),
                   "Natural Language :: English",
                   "Operating System :: OS Independent",
                   "Programming Language :: Python",
                   "Topic :: Scientific/Engineering :: Bio-Informatics"])



'''
Created on June 14, 2018

@author: Stefan.M.Janssen@gmail.com
'''
import sys
import tempfile
import shutil
import unittest
import sepp
from sepp.filemgr import get_data_path
from sepp.config import set_checkpoint
import sepp.scheduler

from sepp.exhaustive import ExhaustiveAlgorithm

# sepp._DEBUG = True
# sepp.reset_loggers()


class Test(unittest.TestCase):
    x = None

    def resetSepp(self):
        # ensure necessary settings are made
        sepp.scheduler._jobPool = None
        sepp.scheduler._parser = None
        self.x = ExhaustiveAlgorithm()
        self.x.options.alignment_file = open(
            get_data_path(
                "q2-fragment-insertion/reference_alignment_tiny.fasta"), "r")
        self.x.options.info_file = open(
            get_data_path(
                "q2-fragment-insertion/RAxML_info-reference-gg-raxml-bl.info"),
            "r")
        self.x.options.tree_file = open(
            get_data_path(
                "q2-fragment-insertion/reference_phylogeny_tiny.nwk"), "r")
        self.x.options.fragment_file = open(
            get_data_path(
                "q2-fragment-insertion/input_fragments.fasta"), "r")
        self.x.options.outdir = tempfile.mkdtemp()
        self.x.options.placement_size = 20

    def setUp(self):
        sys.argv = [sys.argv[0]]
        self.resetSepp()
        self._checkpointfile = tempfile.mktemp()

    def tearDown(self):
        shutil.rmtree(self.x.options.outdir, ignore_errors=True)
        self.resetSepp()

    def test_make_checkpoints(self):
        self.x.options.checkpoint = set_checkpoint(self._checkpointfile)
        self.x.options.checkpoint_interval = 5
        self.x.run()
        self.assertTrue(self.x.results is not None)


if __name__ == "__main__":
    unittest.main()



'''
Created on June 14, 2018

@author: Stefan.M.Janssen@gmail.com
'''
import sys
import tempfile
import shutil
import unittest
import sepp
from sepp.filemgr import get_data_path
import platform
from argparse import Namespace
import sepp.scheduler

from sepp.exhaustive import ExhaustiveAlgorithm
# sepp._DEBUG = True
# sepp.reset_loggers()


class Test(unittest.TestCase):
    x = None

    def resetSepp(self):
        sepp.scheduler._jobPool = None
        self.x = ExhaustiveAlgorithm()
        self.x.options.alignment_file = open(
            get_data_path(
                "q2-fragment-insertion/reference_alignment_tiny.fasta"), "r")
        self.x.options.info_file = open(
            get_data_path(
                "q2-fragment-insertion/RAxML_info-reference-gg-raxml-bl.info"),
            "r")
        self.x.options.tree_file = open(
            get_data_path(
                "q2-fragment-insertion/reference_phylogeny_tiny.nwk"), "r")
        self.x.options.outdir = tempfile.mkdtemp()

        suff_bit = "-64" if sys.maxsize > 2**32 else "-32"
        if platform.system() == 'Darwin':
            suff_bit = ""
        for prog in ['hmmalign', 'hmmbuild', 'hmmsearch', 'pplacer']:
            setattr(self.x.options, prog, Namespace(
                path=get_data_path("../../../tools/bundled/%s/%s%s" % (
                    platform.system(), prog, suff_bit))))

    def setUp(self):
        # ensure necessary settings are made
        sys.argv = [sys.argv[0], "-c", get_data_path("configs/test2.config")]
        self.resetSepp()

    def tearDown(self):
        shutil.rmtree(self.x.options.outdir, ignore_errors=True)
        self.resetSepp()

    def test_diamMid(self):
        self.x.options.fragment_file = open(
            get_data_path(
                "q2-fragment-insertion/input_fragments.fasta"), "r")
        self.x.options.maxDiam = 0.1
        self.x.options.fragmentChunkSize = 1
        self.x.options.decomp_strategy = "midpoint"
        self.x.run()
        self.assertTrue(self.x.results is not None)

    def test_diamCent(self):
        self.x.options.fragment_file = open(
            get_data_path(
                "q2-fragment-insertion/input_fragments.fasta"), "r")
        self.x.options.maxDiam = 0.1
        self.x.options.fragmentChunkSize = 1
        self.x.options.decomp_strategy = "centroid"
        self.x.run()
        self.assertTrue(self.x.results is not None)


if __name__ == "__main__":
    unittest.main()



'''
Created on Sep 19, 2012

@author: smirarab
'''
from sepp.scheduler import Job, JobError
from sepp import get_logger
from abc import abstractmethod
from subprocess import Popen

import os
import shutil
import subprocess
import stat
import re
from sepp.tree import PhylogeneticTree
import sepp.config
import traceback

import io
try:
    filetypes = (io.IOBase, file)
except NameError:
    filetypes = io.IOBase

_LOG = get_logger(__name__)


class ExternalSeppJob(Job):
    '''
    All Sepp jobs that run external programs
    should extend this abstract class.
    This class handles executing external jobs, error handling, and more.
    '''
    def __init__(self, jobtype, path=None, **kwargs):
        Job.__init__(self)
        self.job_type = jobtype
        self._id = None  # _process id for this job
        self._kwargs = dict(kwargs)
        self._process = None
        ''' The following will contain stdout and stderr values *only if* those
        are piped. If not, user should read the files to find the output if
        needed.
        '''
        self.stdoutdata = self.stderrdata = None
        ''' When the following is something other than None, it is piped in to
        the subprocess.
        '''
        self.stdindata = None
        # setting this variable tell JobPoll that errors in this job can be
        # ignored when waiting for reults of all jobs to finish
        self.ignore_error = False
        self.fake_run = False
        self.attributes = dict()
        if path:
            self.path = path
        else:
            self.path = sepp.config.options().__getattribute__(
                self.job_type).path
        self.results_on_temp = False

    def get_id(self):
        return self._id
    id = property(get_id)

    def get_process(self):
        return self._process
    process = property(get_process)

    def run(self):
        ''' Runs the external job, and handles errors, piping, checkpointing,
        etc. get_invocation() needs to be implemented in child classes.
        '''
        # Allow re-execution of finished jobs, useful in checkpointing
        if self.result_set:
            return self.result
        if self.fake_run:
            return self.read_results()
        try:
            _LOG.debug("Starting %s Job with input: %s" % (
                self.job_type, self.characterize_input()))

            assert self.result_set is False, "Job is already run."

            # By default discard standard otuput and error from external
            # programs
            if 'stdout' not in self._kwargs:
                self._kwargs['stdout'] = subprocess.PIPE
            elif isinstance(self._kwargs['stdout'], str):
                self._kwargs['stdout'] = open(self._kwargs['stdout'], 'w')

            if 'stderr' not in self._kwargs:
                self._kwargs['stderr'] = subprocess.PIPE
                # k['stderr'] = open(os.devnull, 'w')
            elif isinstance(self._kwargs['stderr'], str):
                self._kwargs['stderr'] = open(self._kwargs['stderr'], 'w')

            if self.stdindata is not None:
                self._kwargs['stdin'] = subprocess.PIPE

            assert \
                (self.get_invocation()[0].count("/") == 0 or os.path.exists(
                    self.get_invocation()[0])), \
                ("path for %s  does not exist (%s)" %
                 (self.job_type, self.get_invocation()[0]))
            _LOG.debug("Invocation of %s", " ".join(
                (str(x) if x is not None else "?NoneType?"
                 for x in self.get_invocation())))
            self._process = Popen(
                self.get_invocation(), universal_newlines=True, **self._kwargs)
            self._id = self._process.pid

            if self.stdindata is not None:
                (self.stdoutdata, self.stderrdata) = \
                    self._process.communicate(input=self.stdindata)
            else:
                (self.stdoutdata, self.stderrdata) = \
                    self._process.communicate()

    #        if self.process.stdout:
    #            self.process.stdout.close()
    #        if self.process.stderr:
    #            self.process.stderr.close()

    #        except:
    #            raise Exception("\n".join(["Failed to run the following
    #                                        execution:",
    #                                       '" "'.join(self._invocation)])

            if self.process.returncode == 0:
                _LOG.info("Finished %s Job with input: %s" %
                          (self.job_type, self.characterize_input()))
                _LOG.debug(
                    ("Finished %s Job with input: %s with:\n"
                     " return code: %s\n output: %s") % (
                        self.job_type, self.characterize_input(),
                        self.process.returncode,
                        "%s ... (continued: %d ) ..." % (
                            self.stdoutdata[0:100],
                            len(self.stdoutdata)) if self.stdoutdata and len(
                                self.stdoutdata) > 100 else self.stdoutdata))

            else:
                _LOG.debug(
                    "Finished %s Job with input: %s with:\n"
                    " return code: %s\n output: %s\n error:%s"
                    % (self.job_type, self.characterize_input(),
                       self.process.returncode, self.stdoutdata,
                       self.read_stderr()))
                raise JobError("\n".join([
                    "The following execution failed:",
                    ' '.join(self.get_invocation()),
                    self.read_stderr()
                    if self.read_stderr() else 'No error messages available']))

            return self.read_results()
        except Exception:
            traceback.print_exc()
            raise

    def read_stderr(self):
        '''
        Used for reading standard error when an error is detected.
        '''
        if self.stderrdata is not None:
            return self.stderrdata
        elif ("stderr" in self._kwargs and
              isinstance(self._kwargs["stderr"], filetypes)):
            return open(self._kwargs["stderr"].name, 'r').read()
        else:
            return None

    @abstractmethod
    def get_invocation(self):
        '''
        The method needs to return a list with the first argument giving
        the executable, and the rest giving the arguments.
        '''
        raise NotImplementedError(
            "get_invocation should be implemented by subclasses")

    @abstractmethod
    def characterize_input(self):
        '''
        Need to implement this method to help with automatic logging.
        Output a string characterizing the input to this job
        '''
        return ""

    @abstractmethod
    def read_results(self):
        '''
        This method should read the results of an external execution, and turn
        the results into a python object (could be simply the path to an output
        file) and return that python object. This is the result that will be
        ultimately pickled and sent to the main process, and will be accessible
        to the other jobs, joins, etc. Better not to pass around large files.
        This results should be picklable.
        '''
        raise NotImplementedError(
            "read_results should be implemented by subclasses")

    def get_attribute(self, key):
        ''' each job maintains a dictionary of free form attributes.
        '''
        return self.attributes[key]

    def set_attribute(self, key, val):
        ''' each job maintains a dictionary of free form attributes.
        '''
        self.attributes[key] = val


class HMMBuildJob(ExternalSeppJob):
    '''
    The Job class that executes a HMM build
    '''

    def __init__(self, **kwargs):
        self.job_type = "hmmbuild"
        ExternalSeppJob.__init__(self, self.job_type, **kwargs)
        self.infile = None  # input reference alignment
        self.informat = None  # format of input reference alignment
        self.outfile = None  # location of output file
        self.molecule = None  # type of molecule used
        self.symfrac = True  # whether or not symfrac flag is set
        self.options = ""  # no other hmmer options

    def setup(self, infile, outfile, symfrac=True, informat="fasta",
              molecule="dna", **kwargs):
        '''
        Use this to setup the job if you already have input file written to a
        file. Use setup_for_subproblem when possible.
        '''
        self.infile = infile
        self.informat = informat
        self.outfile = outfile
        self.molecule = molecule
        self.symfrac = symfrac
        if ('options' in kwargs):
            self.options = kwargs['options']

    def setup_for_subproblem(self, subproblem, symfrac=True,
                             molecule="dna", **kwargs):
        '''
        Automatically sets up a job given a subproblem object. It outputs the
        right alignment subset to a temporary file.
        '''
        assert isinstance(subproblem, sepp.problem.SeppProblem)
        assert isinstance(
            subproblem.subalignment, sepp.problem.ReadonlySubalignment)
        self.symfrac = symfrac
        # pdb.set_trace()
        self.infile = sepp.filemgr.tempfile_for_subproblem(
            "hmmbuild.input.", subproblem, ".fasta")
        # pdb.set_trace()
        subproblem.write_subalignment_without_allgap_columns(self.infile)
        # pdb.set_trace()
        self.informat = "fasta"
        self.symfrac = symfrac
        self.outfile = sepp.filemgr.tempfile_for_subproblem(
            "hmmbuild.model.", subproblem)
        self.molecule = molecule
        if ('options' in kwargs):
            self.options = kwargs['options']

    def get_invocation(self):
        invoc = [self.path, '--ere', '0.59', "--cpu", "1",
                 "--%s" % self.molecule]
        if self.symfrac is True:
            invoc.extend(["--symfrac", "0.0"])
        if self.options != "":
            invoc.extend(self.options.split())
        if self.informat == "fasta":
            invoc.extend(['--informat', 'afa'])
        invoc.extend([self.outfile, self.infile])
        _LOG.debug("Running HMMBUILD command: %s" % " ".join(invoc))
        return invoc

    def characterize_input(self):
        return self.infile

    def read_results(self):
        '''
        Simply make sure the file exists and is not empty. Don't need to load
        the file into memory or anything else. Just return the location of the
        file.
        '''
        assert os.path.exists(self.outfile)
        assert os.stat(self.outfile)[stat.ST_SIZE] != 0
        return self.outfile


class HMMAlignJob(ExternalSeppJob):
    def __init__(self, **kwargs):
        self.job_type = 'hmmalign'
        ExternalSeppJob.__init__(self, self.job_type, **kwargs)
        self.hmmmodel = None
        self.fragments = None
        self.outfile = None
        self.base_alignment = None
        self.trim = None
        self.molecule = None

    def setup(self, hmmmodel, fragments, output_file, base_alignment=None,
              trim=True, molecule="dna", **kwargs):
        '''
        Setup job parameters when those are externally decided.
        Use setup_for_subproblem when possible.
        '''
        self.hmmmodel = hmmmodel
        self.fragments = fragments
        self.outfile = output_file
        self.base_alignment = base_alignment
        self.trim = trim
        self.molecule = molecule
        self._kwargs = kwargs

    def partial_setup_for_subproblem(self, subproblem,
                                     trim=False, molecule="dna", **kwargs):
        '''Automatically sets up a job given a subproblem object. Note that
        hmmmodel is not set and fragments is just a filename that needs to be
        created later. base_alignment is not set either.
        '''
        assert isinstance(subproblem, sepp.problem.SeppProblem)

        self.outfile = sepp.filemgr.tempfile_for_subproblem(
            "hmmalign.results.", subproblem)
        self.fragments = sepp.filemgr.tempfile_for_subproblem(
            "hmmalign.frag.", subproblem, ".fasta")

        self.trim = trim
        self.molecule = molecule
        self._kwargs = kwargs

    def get_invocation(self):
        invoc = [self.path, "--%s" % self.molecule,
                 "-o", self.outfile]

        if self.trim:
            invoc.extend(["--trim"])
        # if self.base_alignment:
        #    invoc.extend["--mapali" , self.base_alignment]
        if "user_options" in self._kwargs:
            invoc.extend(self._kwargs["user_options"].split())
        invoc.extend([self.hmmmodel, self.fragments])
        return invoc

    def characterize_input(self):
        return "model:%s, fragments:%s, trim:%s, base_alignment:%s" % (
            self.hmmmodel, self.fragments, self.trim, self.base_alignment)

    def read_results(self):
        '''
        Since the output file can be huge, we don't want to read it here,
        because it will need to get pickled and unpickled. Instead, we just
        send back the file name, and will let the caller figure out what to do
        with it.
        '''
        if self.fake_run:
            return None
        if os.path.exists(self.outfile):
            assert os.stat(self.outfile)[stat.ST_SIZE] != 0
            return self.outfile
        else:
            return None


class HMMSearchJob(ExternalSeppJob):

    def __init__(self, **kwargs):
        self.job_type = 'hmmsearch'
        ExternalSeppJob.__init__(self, self.job_type, **kwargs)
        self.hmmmodel = None
        self.fragments = None
        self.outfile = None
        self.elim = None
        self.filters = None
        if hasattr(sepp.config.options().hmmsearch, "piped"):
            self.pipe = sepp.config.options().hmmsearch.piped.\
                strip().lower() == "true"
        else:
            self.pipe = True
        # _LOG.info("Pipe: %s" %str(self.pipe ))
        self.results_on_temp = not self.pipe
        _LOG.debug("HmmSearch: Piped?: %s and keep on temp?: %s" % (
            str(self.pipe), str(self.results_on_temp)))

    def setup(self, hmmmodel, fragments, output_file, elim=None,
              filters=True, **kwargs):
        self.hmmmodel = hmmmodel
        self.fragments = fragments
        self.outfile = output_file
        self.elim = elim
        self.filters = filters
        self._kwargs = kwargs

    def partial_setup_for_subproblem(self, fragments_file, subproblem,
                                     elim=None, filters=True, **kwargs):
        '''
        Automatically sets up a job given a subproblem object.
        Note that hmmmodel is not setup and needs to be set separately.
        '''
        assert isinstance(subproblem, sepp.problem.SeppProblem)

        self.outfile = sepp.filemgr.tempfile_for_subproblem(
            "hmmsearch.results.", subproblem)
        self.fragments = fragments_file
        if not self.fragments:
            self.fake_run = True
        self.elim = elim
        self.filters = filters
        self._kwargs = kwargs

    def get_invocation(self):
        invoc = [self.path, "--noali", "--cpu", "1"]
        if not self.pipe:
            invoc.extend(["-o", self.outfile])
        if self.elim is not None:
            invoc.extend(["-E", str(self.elim)])
        if not self.filters:
            invoc.extend(["--max"])
        if "user_options" in self._kwargs:
            invoc.extend(self._kwargs["user_options"].split())
        invoc.extend([self.hmmmodel, self.fragments])
        return invoc

    def characterize_input(self):
        return "model:%s, fragments:%s, elim:%s, filter:%s, output:%s" % (
            self.hmmmodel, self.fragments, self.elim, self.filters,
            "Piped" if self.pipe else self.outfile)

    def read_results(self):
        '''
           Reads the search output file and returns a dictionary that contains
           the e-values of the searched fragments
        '''
        if self.results_on_temp:
            if self.fake_run:
                res = {}
            else:
                assert os.path.exists(self.outfile)
                assert os.stat(self.outfile)[stat.ST_SIZE] != 0
                with open(self.outfile, 'r') as outfile:
                    res = self.read_results_from_temp(outfile)
            with open(self.outfile, 'w') as target:
                target.write(str(res))
            return self.outfile
        else:
            if self.fake_run:
                res = {}
            elif self.pipe:
                outfile = self.stdoutdata.split("\n")
                res = self.read_results_from_temp(outfile)
            else:
                outfile = open(self.outfile, 'r')
                res = self.read_results_from_temp(outfile)
                outfile.close()
            return res

        # Group 1 (e-value) 2 (bitscore) and 9 (taxon name) contain the
        # relevant information, other ones can be ignored unless we plan to
        # do something later

    def read_results_from_temp(self, outfile):
        results = {}

        pattern = re.compile(
            r"([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+"
            r"([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)")
        start_reading = False
        for line in outfile:
            line = line.strip()
            if (not start_reading and line.startswith("E-value") is True):
                start_reading = True
            elif (start_reading and line == ""):
                start_reading = False
                break
            elif (start_reading):
                matches = pattern.search(line)
                if (matches is not None and matches.group(0).find("--") == -1):
                    results[matches.group(9).strip()] = (
                        float(matches.group(1).strip()),
                        float(matches.group(2).strip()))
                    # _LOG.debug("Fragment scores;"
                    #           "fragment:%s E-Value:%s BitScore:%s" %(matches
                    # .group(9).strip(),matches.group(1).strip(), matches.
                    # group(2).strip()))
        return results


class PplacerJob(ExternalSeppJob):
    def __init__(self, **kwargs):
        self.job_type = 'pplacer'
        # pdb.set_trace()
        ExternalSeppJob.__init__(self, self.job_type, **kwargs)
        '''The following is just a string indicating the type of input provided
        to pplacer job. Different setup methods can setup this job with
        different types of input (i.e. reference package, separate refernce
        alignment, etc.)
        '''
        self.setup_setting = None
        self.tree_file = None
        self.backbone_alignment_file = None
        self.info_file = None
        self.extended_alignment_file = None
        self.full_extended_alignment_file = None
        self.out_file = None

    def setup(self, backbone_alignment_file, tree_file, info_file,
              extended_alignment_file, full_extended_alignment_file,
              output_file, **kwargs):
        # pdb.set_trace()
        self.backbone_alignment_file = backbone_alignment_file
        self.tree_file = tree_file
        self.info_file = info_file
        self.extended_alignment_file = extended_alignment_file
        self.full_extended_alignment_file = full_extended_alignment_file
        self.out_file = output_file
        self._kwargs = kwargs
        self.setup_setting = "File:TrInEx"

    def partial_setup_for_subproblem(self, subproblem, info_file, i, **kwargs):
        ''' Automatically sets up a job given a subproblem object.
        Note that extended alignment and the backbone_alignment_file are just
        file names, referring to empty files at this point. These files needs
        to be created before the job is queued.
        '''
        # pdb.set_trace()
        assert isinstance(subproblem, sepp.problem.SeppProblem)
        self.backbone_alignment_file = sepp.filemgr.tempfile_for_subproblem(
            "pplacer.backbone.", subproblem, ".fasta")
        self.tree_file = sepp.filemgr.tempfile_for_subproblem(
            "pplacer.tree.", subproblem, ".tre")
        self.extended_alignment_file = \
            sepp.filemgr.tempfile_for_subproblem("pplacer.extended.%d." % i,
                                                 subproblem, ".fasta")
        self.full_extended_alignment_file = \
            sepp.filemgr.tempfile_for_subproblem(
                "pplacer.full.extended.%d." % i, subproblem, ".fasta")
        self.out_file = os.path.join(
            sepp.filemgr.tempdir_for_subproblem(subproblem),
            self.extended_alignment_file.replace("fasta", "jplace"))
        assert isinstance(subproblem.subtree, PhylogeneticTree)
        subproblem.subtree.write_newick_to_path(self.tree_file,)

        self.info_file = info_file.name \
            if hasattr(info_file, "name") else info_file
        self._kwargs = kwargs
        self.setup_setting = "File:TrInEx"

    def get_invocation(self):
        invoc = [self.path,
                 "--out-dir", os.path.dirname(self.out_file)]
        if "user_options" in self._kwargs:
            invoc.extend(self._kwargs["user_options"].split())

        if self.setup_setting == "File:TrInEx":
            invoc.extend(["-j", "1",
                          "-r", self.backbone_alignment_file,
                          "-s", self.info_file,
                          "-t", self.tree_file,
                          "--groups", "10",
                          self.extended_alignment_file])
        return invoc

    def characterize_input(self):
        if self.setup_setting == "File:TrInEx":
            return ("backbone_alignment_file:%s, tree_file:%s, info_file:%s, "
                    "extended alignment:%s, output:%s") % (
                        self.backbone_alignment_file, self.tree_file,
                        self.info_file, self.extended_alignment_file,
                        self.out_file)
        else:
            return "Not setup properly"

    def read_results(self):
        '''
        Since the output file can be huge, we don't want to read it here,
        because it will need to get pickled and unpickled. Instead, we just
        send back the file name, and will let the caller figure out what to do
        with it.

        But if it is a fake job, then return nothing
        '''
        if (self.fake_run):
            return None
        assert os.path.exists(self.out_file)
        assert os.stat(self.out_file)[stat.ST_SIZE] != 0
        return self.out_file


class MergeJsonJob(ExternalSeppJob):
    def __init__(self, **kwargs):
        self.job_type = 'jsonmerger'
        ExternalSeppJob.__init__(self, self.job_type, **kwargs)
        self.input_string = None
        self.out_file = None

    def setup(self, inString, output_file, **kwargs):
        self.stdindata = inString
        self.out_file = output_file
        self._kwargs = kwargs

    def get_invocation(self):
        invoc = ["java", "-jar", self.path,
                 "-", "-", self.out_file]
        return invoc

    def characterize_input(self):
        return "input:pipe output:%s" % (self.out_file)

    def read_results(self):
        '''
        Since the output file can be huge, we don't want to read it here,
        because it will need to get pickled and unpickled. Instead, we just
        send back the file name, and will let the caller figure out what to do
        with it.
        '''
        assert os.path.exists(self.out_file)
        assert os.stat(self.out_file)[stat.ST_SIZE] != 0
        return self.out_file


class MafftAlignJob(ExternalSeppJob):
    '''
    The Job class that generates a Mafft alignment
    '''

    def __init__(self, **kwargs):
        self.job_type = "mafft"
        ExternalSeppJob.__init__(self, self.job_type, **kwargs)
        self.sequences = None  # input sequences
        self.size = None  # size of backbone
        self.output = None  # output file
        self.threads = 1  # number of threads

    def setup(self, sequences, size, output, threads, **kwargs):
        '''
        Use this to setup the job if you already have input file written to a
        file. Use setup_for_subproblem when possible.
        '''
        self.sequences = sequences
        self.output = output
        self.size = size
        self.threads = threads
        self._kwargs = kwargs

    def setup_for_subproblem(self):
        '''
        Use setup for generating backbone alignment
        '''
        return

    def get_invocation(self):
        invoc = [self.path]
        if (self.size > 200):
            invoc.extend(['--parttree', '--retree', '2', '--partsize', '1000'])
        else:
            invoc.extend(['--localpair', '--maxiterate', '1000'])
        invoc.extend(['--anysymbol', '--thread', str(self.threads),
                      self.sequences])
        return invoc

    def characterize_input(self):
        return "mafftalign %s %s" % (self.sequences, self.output)

    def read_results(self):
        '''
        Read from standard out, write to file
        '''
        output = (self.stdoutdata)
        outfile = open(self.output, 'w')
        outfile.write(output)
        outfile.close()
        return self.output


class FastTreeJob(ExternalSeppJob):
    '''
    The Job class that generates a fasttree tree from an alignment
    '''
    def __init__(self, **kwargs):
        self.job_type = "fasttree"
        ExternalSeppJob.__init__(self, self.job_type, **kwargs)
        self.alignment_file = None  # input alignment
        self.output_file = None  # output tree
        self.molecule = None  # type of molecule

    def setup(self, alignment_file, output_file, molecule, **kwargs):
        '''
        Use this to setup the job if you already have input file written to a
        file. Use setup_for_subproblem when possible.
        '''
        self.alignment_file = alignment_file
        self.output_file = output_file
        self.molecule = molecule
        self._kwargs = kwargs

    def setup_for_subproblem(self):
        '''
        Use setup for generating backbone tree
        '''
        return

    def get_invocation(self):
        invoc = [self.path, '-gtr']
        if (self.molecule != 'protein'):
            invoc.extend(['-nt'])
        invoc.extend(['-quiet', '-nosupport', '-out', self.output_file,
                      self.alignment_file])
        return invoc

    def characterize_input(self):
        return "fasttree %s %s %s" % (
            self.alignment_file, self.output_file, self.molecule)

    def read_results(self):
        '''
        Check and return file
        '''
        assert os.path.exists(self.output_file)
        assert os.stat(self.output_file)[stat.ST_SIZE] != 0
        return self.output_file


class PastaAlignJob(ExternalSeppJob):
    '''
    The Job class that generates a Pasta alignment and tree
    '''
    def __init__(self, **kwargs):
        self.job_type = "pasta"
        ExternalSeppJob.__init__(self, self.job_type, **kwargs)
        self.alignment = None  # input alignment
        self.size = None  # size of backbone
        self.molecule = None  # type of molecule
        self.output = None
        self.config = None
        self.threads = None

    def setup(self, alignment, size, molecule, threads, **kwargs):
        '''
        Use this to setup the job if you already have input file written to
        a file. Use setup_for_subproblem when possible.
        '''
        self.alignment = alignment
        self.size = size
        self.molecule = molecule
        if (molecule == 'aa'):
            self.molecule == 'protein'
        self.threads = threads
        self._kwargs = kwargs

    def setup_for_subproblem(self, subproblem, molecule="dna", **kwargs):
        '''
        Use setup for generating backbone tree
        '''
        return

    def get_invocation(self):
        invoc = [self.path, '--num-cpus=%d' % self.threads, '-i',
                 self.alignment, "--datatype=%s" % self.molecule,
                 '--temporaries=%s/pastaout/' %
                 sepp.filemgr.get_root_temp_dir(),
                 '-j', 'pastajob', '--output-directory=%s/pastaout/' %
                 sepp.filemgr.get_root_temp_dir()]
        return invoc

    def characterize_input(self):
        return " ".join(self.get_invocation())

    def read_results(self):
        '''
        Read the PASTA log file and get the alignment and tree from file
        '''
        assert os.path.exists(
            '%s/pastaout/pastajob.out.txt' % sepp.filemgr.get_root_temp_dir())
        assert os.stat(
            '%s/pastaout/pastajob.out.txt' %
            sepp.filemgr.get_root_temp_dir())[stat.ST_SIZE] != 0
        outfile = open(
            '%s/pastaout/pastajob.out.txt' % sepp.filemgr.get_root_temp_dir(),
            'r')
        alignment_pattern = re.compile('Writing resulting alignment to (.*)')
        tree_pattern = re.compile('Writing resulting tree to (.*)')
        tree_file = ''
        alignment_file = ''
        for line in outfile:
            line = line.strip()
            result = alignment_pattern.findall(line)
            if (len(result) != 0):
                alignment_file = result[0]
            result = tree_pattern.findall(line)
            if (len(result) != 0):
                tree_file = result[0]
        return (alignment_file, tree_file)


class SateAlignJob(ExternalSeppJob):
    '''
    The Job class that generates a Sate alignment and tree
    '''
    def __init__(self, **kwargs):
        self.job_type = "sate"
        ExternalSeppJob.__init__(self, self.job_type, **kwargs)
        self.alignment = None  # input alignment
        self.size = None  # size of backbone
        self.molecule = None  # type of molecule
        self.output = None
        self.config = None
        self.threads = None

    def setup(self, alignment, size, output, molecule, threads, **kwargs):
        '''
        Use this to setup the job if you already have input file written to
        a file. Use setup_for_subproblem when possible.
        '''
        self.alignment = alignment
        self.size = size
        self.output = output
        self.molecule = molecule
        if (molecule == 'aa'):
            self.molecule == 'protein'
        self.threads = threads
        self._kwargs = kwargs

    def setup_for_subproblem(self, subproblem, molecule="dna", **kwargs):
        '''
        Use setup for generating backbone tree
        '''
        return

    def get_invocation(self):
        size_str = '--max-subproblem-size=200'
        if (self.size <= 200):
            size_str = '--max-subproblem-frac=0.50'
        mergerType = 'opal'
        if (self.molecule == 'amino'):
            mergerType = 'muscle'
        invoc = [self.path, '-i', self.alignment, '--merger=%s' % mergerType,
                 '--aligner=mafft', '--tree-estimator=fasttree',
                 '--num-cpus=%d' % self.threads,
                 "--datatype=%s" % self.molecule,
                 '--temporaries=%s/satetmp' % self.output,
                 '--break-strategy=centroid', size_str,
                 '--time-limit=-1', '--iter-without-imp-limit=1', '-j',
                 'satejob',
                 '--output-directory=%s/sateout/' %
                 sepp.filemgr.get_root_temp_dir()]
        return invoc

    def characterize_input(self):
        return " ".join(self.get_invocation())

    def read_results(self):
        '''
        Read the Sate log file and get the alignment and tree from file,
        copy to output directory
        '''
        assert os.path.exists(
            '%s/sateout/satejob.out.txt' % sepp.filemgr.get_root_temp_dir())
        assert os.stat(
            '%s/sateout/satejob.out.txt' %
            sepp.filemgr.get_root_temp_dir())[stat.ST_SIZE] != 0
        outfile = open(
            '%s/sateout/satejob.out.txt' %
            sepp.filemgr.get_root_temp_dir(), 'r')
        alignment_pattern = re.compile('Writing resulting alignment to (.*)')
        tree_pattern = re.compile('Writing resulting tree to (.*)')
        tree_file = ''
        alignment_file = ''
        for line in outfile:
            line = line.strip()
            result = alignment_pattern.findall(line)
            if (len(result) != 0):
                alignment_file = result[0]
            result = tree_pattern.findall(line)
            if (len(result) != 0):
                tree_file = result[0]
        shutil.copyfile(tree_file, "%s/sate.fasttree" % self.output)
        shutil.copyfile(alignment_file, "%s/sate.fasta" % self.output)
        return (tree_file, alignment_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
###########################################################################
#    Copyright 2012 Siavash Mirarab, Nam Nguyen, and Tandy Warnow.
#    This file is part of SEPP, and some parts of the code are
#    adopted from SATe software package
#   (https://github.com/sate-dev/sate-core/blob/master/sate/alignment.py).
#
#    SEPP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SEPP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SEPP.  If not, see <http://www.gnu.org/licenses/>.
###########################################################################

import re
# from sepp import #_LOG
from sepp.filemgr import open_with_intermediates

from collections import Mapping
import copy
from sepp import get_logger
import io
try:
    filetypes = (io.IOBase, file)
except NameError:
    filetypes = io.IOBase


_INDEL = re.compile(r"[-]")
_DANGEROUS_NAME_CHARS = re.compile(r"[^a-zA-Z0-9]")

DATASET_TAXA_ATTR = "taxon_sets"
DATASET_CHAR_ATTR = "char_matrices"

_LOG = get_logger(__name__)


def get_pdistance(distances, leaves, stat='mean'):
    """Returns the mean p-distance given a distance matrix and a set of
       names"""
    counts = 0
    pdistance = 0
    for i in range(0, len(leaves)-1):
        for j in range(i+1, len(leaves)):
            name = "".join([leaves[i], leaves[j]])
            if stat == 'mean':
                pdistance = pdistance + distances[name]
                counts = counts + 1
            elif stat == 'max':
                if (distances[name] > pdistance):
                    pdistance = distances[name]

    if (stat == 'mean'):
        pdistance = pdistance / counts
    return pdistance


def hamming_distance(seq1, seq2):
    """Returns hamming distance between two sequences"""
    # xors = [ord(a) ^ ord(b) for a,b in zip(seq1,seq2)]
    # ors = [ord(a) | ord(b) for a,b in zip(seq1,seq2)]
    # num_mismatch = 0.0
    # non_gap = 0.0
    # for (a,b) in zip(xors,ors):
    #     if (a != 0 & a <= 127):
    #         num_mismatch+=1
    #     if (b != 255):
    #         non_gap+=1
    #
    # if non_gap == 0:
    #     return -1
    # return num_mismatch/non_gap;
    length = len(seq1)
    num_mismatch = 0.0
    non_gap = 0.0
    for i in range(0, length):
        if seq1[i] == "-" or seq2[i] == "-":
            continue
        non_gap += 1
        if seq1[i].lower() != seq2[i].lower():
            num_mismatch += 1
    if non_gap == 0:
        return 0
    return num_mismatch / non_gap


def is_sequence_legal(seq):
    """Check for illegal characters -- TODO:, currently returns True"""
    return True


def _read_fasta(src):
    """generator that returns (name, sequence) tuples from either a FASTA
    formatted file or file object.
    Use MutableAlignment class instead of directly using this method
    """
    file_obj = None
    if isinstance(src, str):
        try:
            file_obj = open(src, "rU")
        except IOError:
            print(("The file `%s` does not exist, exiting gracefully" % src))
    elif isinstance(src, filetypes):
        file_obj = src
    else:
        raise TypeError('FASTA reader cannot recognize the source of %s' % src)
    name = None
    seq_list = list()
    for line_number, i in enumerate(file_obj):
        if i.startswith('>'):
            if name:
                yield name, ''.join(seq_list)
                seq_list = list()
            name = i[1:].strip()
        else:
            seq = ''.join(i.strip().upper().split())
            if not is_sequence_legal(seq):
                raise Exception(
                    "Error: illegal characeters in sequence at line %d" %
                    line_number)
            seq_list.append(seq)
    yield name, ''.join(seq_list)
    if isinstance(src, str):
        file_obj.close()


def _write_fasta(alignment, dest):
    """Writes the `alignment` in FASTA format to either a file object or
       file"""
    file_obj = None
    if isinstance(dest, str):
        file_obj = open(dest, "w")
    else:
        file_obj = dest
    for name in sorted(alignment.keys()):
        file_obj.write('>%s\n%s\n' % (name, alignment[name]))
    if isinstance(dest, str):
        file_obj.close()

# def is_taxon_a_full_length_seq(x):
#    if isinstance(x, str):
#        return False if "_" in x else True
#    return False if " " in x.label else True

# def get_taxon_label(taxon):
#    return taxon.label.replace(" ", "_")


class ReadOnlyAlignment(Mapping, object):
    '''
    An abstract class that provide all the read-only operations for an
    alignment'''
    def get_num_taxa(self):
        """returns the number sequences"""
        return len(self.get_sequence_names())

    def get_p_distance(self):
        """returns the average and max p-distance of alignment"""
        if self.is_empty():
            raise ValueError("The alignment is empty.\n")
        max = 0
        average = 0.0
        count = 0.0
        copy = MutableAlignment()
        copy.set_alignment(self)
        names = copy.get_sequence_names()
        for name, val in copy.items():
            copy[name] = val.replace('-', chr(255)).lower()

        for x in range(0, self.get_num_taxa()):
            for y in range(x+1, self.get_num_taxa()):
                distance = hamming_distance(copy[names[x]], copy[names[y]])
                if distance >= 0:
                    if distance > max:
                        max = distance
                    count += 1
                    average += distance
        average = average / count
        return (average, max)

    def get_length(self):
        """returns the number of columns in the alignment"""
        if self.is_empty():
            raise ValueError("The alignment is empty.\n")
        else:
            return len(next(iter(self.values())))

    def get_sequence_names(self):
        """returns a list of sequence names"""
        return list(self.keys())

    def write_to_path(self, filename, schema='FASTA'):
        """Writes the sequence data in the specified `file_format` to
           `filename`"""
        _LOG.debug(
            "Writing alignment of type %s with length %d to file %s" %
            (str(self.__class__), len(self), filename))
        file_obj = open_with_intermediates(filename, 'w')
        self.write(file_obj, file_format=schema)
        file_obj.close()

    def write(self, file_obj, file_format):
        """Writes the sequence data in the specified `file_format` to
           `file_obj`"""
        if (file_format.upper() == 'FASTA'):
            write_func = _write_fasta
#        elif (file_format.upper() == 'NEXUS'):
#            write_func = write_nexus
#        elif (file_format.upper() == 'PHYLIP'):
#            write_func = write_phylip
        else:
            write_func = _write_fasta
        write_func(self, file_obj)

    def write_unaligned_fasta(self, filename):
        """Writes the sequence data without gaps as FASTA, but note that the
        lines may bet "ragged".
        """
        file_obj = open_with_intermediates(filename, 'w')
        for name, seq in self.items():
            file_obj.write('>%s\n%s\n' % (name, re.sub(_INDEL, '', seq)))
        file_obj.close()

    def is_empty(self):
        return self.__len__() < 1

    def is_aligned(self):
        if self.is_empty():
            raise ValueError("The alignment is empty.\n")
        else:
            l0 = self.get_length()
            return all(len(i) == l0 for i in self.values())

    def is_all_gap(self, pos):
        '''Checks to see if a column is all gap column at position x'''
        for seq in self.values():
            if (seq[pos] != '-'):
                return False
        return True

    def __str__(self):
        return '\n'.join([">%s\n%s" % (k, self[k])
                          for k in sorted(self.keys())])

    def divide_to_equal_chunks(self, chunks, max_chunk_size=None):
        names = self.get_sequence_names()
        ret = []
        if max_chunk_size and len(names) / chunks > max_chunk_size:
            chunks = len(names) // max_chunk_size + 1
        for i in range(0, chunks):
            subset = names[i:len(names):chunks]
            if subset:
                subset_alg = self.get_soft_sub_alignment(subset)
            else:
                subset_alg = None
            ret.append(subset_alg)
        return ret

    def get_soft_sub_alignment(self, sub_key):
        '''
        Returns a read-only sub-alignment, which won't consume extra memory,
        since it will not hold a separate copy of the alignment.
        '''
        return ReadonlySubalignment(sub_key, self)


class MutableAlignment(dict, ReadOnlyAlignment, object):
    """ An alignment object, that can be modified. This is the class that
    should be used mainly for holding alignments.
    """
    def __init__(self):
        "creates an empty matrix"
        dict.__init__(self)
        self.datatype = None

    def set_alignment(self, alignment):
        for name, seq in alignment.items():
            self[name] = seq.upper()

    def read_filepath(self, filename, file_format='FASTA'):
        """Augments the matrix by reading the filepath.
        If duplicate sequence names are encountered then the old name will
        be replaced.
        """
        file_obj = open(filename, 'rU')
        return self.read_file_object(file_obj, file_format=file_format)

    def read_file_object(self, file_obj, file_format='FASTA'):
        """Augments the matrix by reading the file object.
        If duplicate sequence names are encountered then the old name will
        be replaced.
        """
        if (file_format.upper() == 'FASTA'):
            read_func = _read_fasta
#        elif (file_format.upper() == 'NEXUS'):
#            read_func = read_nexus
#        elif (file_format.upper() == 'PHYLIP'):
#            read_func = read_phylip
        else:
            raise NotImplementedError(
                "Unknown file format (%s) is not supported" % file_format)
        for name, seq in read_func(file_obj):
            self[name] = seq.upper()
        return self

    def add_column(self, pos, char='-'):
        # _LOG.debug("Added a column to reference alignment at position
        # %d" %pos)
        for name, seq in self.items():
            if hasattr(char, "get"):
                c = char.get(name, '-')
            else:
                c = char
            seq = seq[:pos] + c + seq[pos:]
            self[name] = seq

    def remove_column(self, pos):
        for name, seq in self.items():
            seq = seq[:pos] + seq[pos + 1:]
            self[name] = seq

    def remove_columns(self, indexes):
        for name, seq in self.items():
            self[name] = ''.join((
                char for idx, char in enumerate(seq) if idx not in indexes))

    def delete_all_gap(self):
        '''
        Delete all sites that consists of nothing but gaps
        '''
        # pdb.set_trace()
        pos = 0
        i = 0
        subset = []
        while (pos < self.get_length()):
            if (self.is_all_gap(pos)):
                self.remove_column(pos)
            else:
                subset.append(i)
                pos = pos + 1
            i += 1

        _LOG.debug("Alignment length reduced to %d" % len(subset))
        return subset

    def degap(self):
        '''
        remove all
        '''
        for name, seq in self.items():
            self[name] = self[name].replace("-", "")

    def get_hard_sub_alignment(self, sub_keys):
        "Creates a new alignment with a subset of the taxa."
        new_alignment = MutableAlignment()
        new_alignment.datatype = self.datatype
        for key in sub_keys:
            new_alignment[key] = self[key]
        new_alignment.delete_all_gap()
        return new_alignment


class ReadonlySubalignment(ReadOnlyAlignment):
    '''
    A class that emulates a subalignment of a given alignment. This is a
    readonly alignment and does not actually hold sequences in memory. It
    simply keeps a set of sequences that belong to the sub alignment,
    and implements methods of alignment class (and dictionaries) such that
    only subalignment keys are returned.
    '''
    def __init__(self, keys, parent_alignment):
        self.seq_names = set(keys)
        self.parent_alignment = parent_alignment

    def __getitem__(self, key):
        if key in self.seq_names:
            return self.parent_alignment[key]
        else:
            raise KeyError(key)

    def __len__(self):
        return len(self.seq_names)

    def __iter__(self):
        for key in self.seq_names:
            yield key

    def get_mutable_alignment(self):
        ret = MutableAlignment()
        ret.set_alignment(self)
        return ret


class _AlignmentLookupHelper(object):
    '''
    Internal helper calss
    '''
    def __init__(self, pos, ref):
        self.pos = pos
        self.ref = ref

    def get(self, key, default=None):
        if key in self.ref:
            return self.ref[key][self.pos]
        else:
            return default


class ExtendedAlignment(MutableAlignment):
    '''
    This is used to keep an extended alignment. An extended alignment
    has a bunch of sequences that are labeled as fragments. More importantly,
    columns of extended alignments are labeled with numbers and also it is
    known whether a column is an "insertion" column or a normal column.
    1) these can be read from .sto files.
    2) these alignments can be merged together.
    '''
    def __init__(self, fragment_names):
        MutableAlignment.__init__(self)
        self.fragments = set(fragment_names)
        self._col_labels = []

    def set_alignment(self, alignment):
        MutableAlignment.set_alignment(self, alignment)
        self._reset_col_names()

    def read_file_object(self, file_obj, file_format='FASTA'):
        ''' currently supports only fasta'''
        ret = MutableAlignment.read_file_object(self, file_obj, file_format)
        self._reset_col_names()
        return ret

    def remove_missing_fragments(self):
        for frag in list(self.get_fragment_names()):
            if frag not in self:
                self.fragments.remove(frag)

    def add_column(self, pos, char='-', new_label=None):
        '''
        A new column is added to the alignment. The new label can be value,
        or one of three directives (see below).
        '''
        MutableAlignment.add_column(self, pos, char)
        if new_label == "MAX":
            self._col_labels.insert(pos, max(self._col_labels) + 1)
        elif new_label == "INC_LAST":
            self._col_labels.add(max(self._col_labels) + 1)
        elif new_label == "RESET":
            self._reset_col_names()
        else:
            self._col_labels.insert(pos, new_label)

    def remove_column(self, pos, labels="REMOVE"):
        '''
        Remove a column and potentially adjust column names.
        '''
        MutableAlignment.remove_column(self, pos)
        if labels == "RESET":
            self._reset_col_names()
        elif labels == "REMOVE":
            self._col_labels = self._col_labels[:pos] + \
                self._col_labels[pos + 1:]

    def _get_col_labels(self):
        return self._col_labels
    col_labels = property(_get_col_labels)

    def _reset_col_names(self):
        ''' sequentially label columns'''
        self._col_labels = list(range(0, self.get_length()))

    def get_fragments_readonly_alignment(self):
        '''
        Return a readonly alignment that contains only the fragments.
        '''
        return ReadonlySubalignment(self.fragments, self)

    def get_base_readonly_alignment(self):
        '''
        Returns a readonly subalignment that does not contain fragments.
        '''
        return ReadonlySubalignment(self.get_base_seq_names(), self)

    def get_fragment_names(self):
        return self.fragments

    def get_base_seq_names(self):
        return list(set(self.keys()) - self.fragments)

    def _read_sto(self, handle):
        '''
        Reads a sto file and populates this current object. Figures out
        insertion columns by finding lower case letters, asterisks, and dots.
        '''
        p = re.compile(r'[a-z*.]')
        insertions = []
        lastStartInd = -1
        for line in handle:
            line = line.strip()
            if line == "//":
                # End of the alignment. ignore meta-data
                break
            elif line == "" or line[0] == "#":
                pass
            else:  # not a comment
                key, seq = line.split()
                current = self.get(key, "")
                startind = len(current)
                # finding insertion columns in limited to first sequence
                if startind != lastStartInd:
                    # s = sum(len(x) for x in current)
                    insertions.extend([m.start() + startind
                                       for m in p.finditer(seq)])
                    lastStartInd = startind
                self[key] = current + seq.replace(".", "-")
#        for k,v in self.items():
#            self[k] = "".join(v)
        self._reset_col_names()
        return set(insertions)

    def build_extended_alignment(self, base_alignment, path_to_sto_extension,
                                 convert_to_string=True):
        '''
        Given a base alignment and a path to an .sto file (or a list of paths),
        this methods populates self with an extended alignment by first reading
        the base alignment, and then merging in all the .sto extension
        alignments.
        Note that the .sto alignments should contain only fragments, and also
        note that there should be no taxon overlap among extensions, or between
        extensions and the base.
        '''
        if isinstance(base_alignment, ReadOnlyAlignment):
            self.set_alignment(copy.deepcopy(base_alignment))
        elif isinstance(base_alignment, str):
            _LOG.debug("Reading base alignment: %s." % (base_alignment))
            self.read_filepath(base_alignment, "FASTA")

        if isinstance(path_to_sto_extension, str):
            paths = [path_to_sto_extension]
        else:
            paths = path_to_sto_extension

        self.from_string_to_bytearray()
        for path in paths:
            _LOG.debug("Reading sto extended alignment: %s." % (path))
            ext = ExtendedAlignment(self.fragments)
            ext.read_extended_alignment(path)
            _LOG.info(
                "Merging extension sto file (%s) into base alignment (%s)." %
                (path, base_alignment))
            self.merge_in(ext, False)
            _LOG.debug(
                ("Finished merging extension sto file (%s) into base "
                 "alignment (%s).") %
                (path, base_alignment))
            del ext
        if convert_to_string:
            self.from_bytearray_to_string()

    def read_extended_alignment(self, path, aformat="stockholm",
                                assertion=False):
        ''' Reads alignment from given path and figures out "insertion"
        columns. Labels insertion columns with special labels and labels the
        rest of columns (i.e. original columns) sequentially.
        '''
        handle = open(path, 'rU')
        insertions = None
        if aformat.lower() == "stockholm":
            insertions = self._read_sto(handle)
            _LOG.debug("%s insertions: %d" % (path, len(insertions)))
        else:
            raise ValueError("format %s is not supported yet" % aformat)
        assert insertions is not None

        '''Assert that insertion columns have only gaps in original seqs and
        give them appropriate labels'''
        insertion = -1
        for c in insertions:
            if assertion:
                for k in self.get_base_seq_names():
                    assert not self[k][c] != "-", \
                        ("Insertion column has sequence among original "
                         "sequences. An error? column: %d k= %s" % (c, k))
            self.col_labels[c] = insertion
            insertion -= 1
        ''' Adjust labels of other columns '''
        i = 0
        for c in range(0, self.get_length()):
            if self.col_labels[c] >= 0:
                self.col_labels[c] = i
                i += 1

    def _is_insertion_label(self, i):
        ''' This differentiates between an insertion and an original column
        '''
        return i < 0

    def is_insertion_column(self, col):
        return self._is_insertion_label(self.col_labels[col])

    def relabel_original_columns(self, original_labels):
        '''
        This methods relabels non-insertion columns in self based on the
        input labels. Insertion column labels will not be affected.
        '''
        j = 0
        _LOG.debug(
            "Relabeling %d (%d) with %d labels." %
            (self.get_length(), len(self._col_labels), len(original_labels)))
        for i in range(0, self.get_length()):
            if not self._is_insertion_label(self.col_labels[i]):
                assert \
                    j < len(original_labels), \
                    ("Not enough labels %d %d %d.\n %s" %
                        (i, j, len(original_labels), str(self._col_labels)))
                self.col_labels[i] = original_labels[j]
                j += 1
        assert\
            j == len(original_labels), \
            ("Some of original labels are unused."
             " Some columns from original alignment went missing? %d %d" %
                (j, len(original_labels)))

    def get_insertion_columns(self):
        return [i
                for (i, x)
                in enumerate(self.col_labels)
                if self._is_insertion_label(x)]

    def get_insertion_column_ranges(self):
        pos = []
        strt = None
        prev = None
        for p in self.get_insertion_columns():
            if prev is None or prev + 1 != p:
                if strt is not None:
                    pos.append((strt, prev))
                strt = p
            prev = p
        if strt is not None:
            pos.append((strt, prev))
        return pos

    def write_insertion_column_indexes(self, path):
        file_obj = open(path, 'w')
        r = ','.join("%d-%d" % (pair[0], pair[1])
                     for pair in self.get_insertion_column_ranges())
        file_obj.write(r)
        file_obj.write("\n")
        file_obj.close()

    def remove_insertion_columns(self):
        '''
        Outputs a new alignment with insertion columns masked out.
        '''
        cols = self.get_insertion_columns()
        s = []
        a = 0
        for b in cols:
            if b > a:
                s.append((a, b))
            a = b + 1
        s.append((a, len(self.col_labels)))
        for name, seq in list(self.items()):
            news = []
            for c in s:
                news.append(seq[c[0]:c[1]])
            self[name] = "".join(news)

    def write_insertion_maked_to_file(self, path):
        cols = self.get_insertion_columns()
        s = []
        a = 0
        for b in cols:
            if b > a:
                s.append((a, b))
            a = b + 1
        s.append((a, len(self.col_labels)))
        file_obj = open(path, 'w')
        for name, seq in self.items():
            file_obj.write('>%s\n' % name)
            for c in s:
                file_obj.write(seq[c[0]:c[1]])
            file_obj.write("\n")
        file_obj.close()

    def from_bytearray_to_string(self):
        for k, v in self.items():
            self[k] = v.decode()

    def from_string_to_bytearray(self):
        for k, v in self.items():
            self[k] = bytearray(v, encoding="utf8")

    def merge_in(self, other, convert_to_string=True):
        '''
        Merges another alignment in with the current alignment. The other
        alignment needs to be an ExtendedAlignment as well. Since both
        alignments are extended alignments, we know column labels, and we also
        know which columns are insertions. Columns with the same labels are
        merged together. In other cases gaps are introduced as necessary to
        merge the two alignments. Insertion columns are considered different
        even when they have the same label.

        convert_to_string is by default true, meaning that in the beginning
        sequences of self are assumed to be string objects. These are converted
        to bytearray for better speed, but at the end of the merge, are
        converted back to string. When convert_to_string is False, no
        conversion back and from string is performed, *but* everything is
        assumed to be in bytearrays.
        So, self is assumed to be in bytearray before calling merge, and will
        remain in bytearray after calling merge. This is useful in cases where
        multiple alignments are merged in with self. Initially, everything
        should be turned into bytearray, and after all merging is finished,
        everything can be converted back to string (using
        from_bytearray_to_string and from_string_to_bytearray).
        '''
        assert isinstance(other, ExtendedAlignment)
        _LOG.debug("Merging started ...")
        if other.is_empty():
            return
        me = 0
        she = 0  # Assumption: alignments are female!
        me_len = self.get_length() if not self.is_empty() else 0
        she_len = other.get_length()
        insertion = -1

        merged_insertion_columns = 0

        ''' Add sequences from her to my alignment '''
        for f in other.fragments:
            self.fragments.add(f)
        if convert_to_string:
            self.from_string_to_bytearray()

        selfother = {}
        for k, v in other.items():
            # assert(k not in self,
            # "Merging overlapping alignments not implemented")
            if (k not in self):
                selfother[k] = bytearray(v, encoding="utf8")
        while True:
            ''' Check exit conditions'''
            if me == me_len and she == she_len:
                break

            ''' Check the 5 possible statuses between she and I '''
            if she != she_len and other.is_insertion_column(she):
                if me != me_len and self.is_insertion_column(me):
                    ''' We both have a series of insertion columns'''
                    start = me
                    while(me != me_len and self.is_insertion_column(me) and
                          she != she_len and other.is_insertion_column(she)):
                        me += 1
                        she += 1
                        merged_insertion_columns += 1
                    run = me - start
                    self.col_labels[start:me] = list(range(
                        insertion, insertion-run, -1))
                else:
                    ''' Hers is a series of insertion columns'''
                    start = she
                    while she != she_len and other.is_insertion_column(she):
                        she += 1
                    run = she - start
                    ins = bytearray(b"-") * run
                    for seq in self.values():
                        seq[me:me] = ins
                    self._col_labels[me:me] = list(range(
                        insertion, insertion - run, -1))
                    insertion -= run
                    me += run
                    me_len += run
            elif me != me_len and self.is_insertion_column(me):
                ''' Mine is a series of insertion column'''
                start = me
                while me != me_len and self.is_insertion_column(me):
                    me += 1
                run = me - start
                ins = bytearray(b"-") * run
                for v in selfother.values():
                    v[start:start] = ins
                self.col_labels[start:me] = list(
                    range(insertion, insertion-run, -1))
                insertion -= run
            elif(she == she_len or (me != me_len and
                 self.col_labels[me] < other.col_labels[she])):
                ''' My column is not present (i.e. was allgap) in the
                    "other"'''
                start = me
                while(me < me_len and (she == she_len or me != me_len and
                      self.col_labels[me] < other.col_labels[she])):
                    me += 1
                run = me - start
                ins = bytearray(b"-") * run
                for v in selfother.values():
                    v[start:start] = ins
            elif(me == me_len or (she != she_len and
                 self.col_labels[me] > other.col_labels[she])):
                ''' Her column is not present (i.e. was allgap) in "me"'''
                start = she
                while(she < she_len and (me == me_len or she != she_len and
                      self.col_labels[me] > other.col_labels[she])):
                    she += 1
                run = she - start
                ins = bytearray(b"-") * run
                for seq in self.values():
                    seq[me:me] = ins
                self._col_labels[me:me] = other.col_labels[start:she]
                me += run
                me_len += run
            elif self.col_labels[me] == other.col_labels[she]:
                ''' A shared column'''
                while(me < me_len and she < she_len and
                      self.col_labels[me] == other.col_labels[she]):
                    she += 1
                    me += 1
            else:
                raise "hmmm, we thought this should be impossible? %d %d" % (
                    me, she)

        self.update(selfother)

        if convert_to_string:
            self.from_bytearray_to_string()
        _LOG.debug("Merging finished ...")

        return merged_insertion_columns
